from __future__ import unicode_literals

import datetime
from itertools import chain

from django.core import validators
from django.db import transaction
from passlib.hash import sha512_crypt
from rest_framework import serializers, exceptions
from waldur_ansible.common import serializers as common_serializers
from waldur_ansible.python_management import serializers as python_management_serializers, utils as pythhon_management_utils, models as python_management_models

from waldur_core.core import models as core_models, serializers as core_serializers
from waldur_core.structure import permissions as structure_permissions, serializers as structure_serializers
from . import models, jupyter_hub_management_service

REQUEST_TYPES_PLAIN_NAMES = {
    models.JupyterHubManagement: 'overall',
    models.JupyterHubManagementSyncConfigurationRequest: 'sync_configuration',
    models.JupyterHubManagementMakeVirtualEnvironmentGlobalRequest: 'globalize_virtual_envs',
    models.JupyterHubManagementMakeVirtualEnvironmentLocalRequest: 'localize_virtual_envs',
    models.JupyterHubManagementDeleteRequest: 'delete_jupyter_hub',
}


class JupyterHubUserSerializer(core_serializers.AugmentedSerializerMixin, serializers.HyperlinkedModelSerializer):
    admin = serializers.BooleanField()
    whitelisted = serializers.BooleanField()
    username = serializers.CharField(max_length=255, validators=[
        validators.RegexValidator(
            regex='^[a-zA-Z0-9_]+$',
            message=b'Username may contain only numbers, characters and underscores!',
        ),
    ])

    class Meta(object):
        model = models.JupyterHubUser
        fields = ('uuid', 'admin', 'whitelisted', 'username', 'password',)
        read_only_fields = ('uuid', 'output', 'state', 'created', 'modified',)
        write_only_fields = ('password',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class JupyterHubManagementRequestMixin(core_serializers.AugmentedSerializerMixin, serializers.HyperlinkedModelSerializer):
    request_type = serializers.SerializerMethodField()
    state = serializers.SerializerMethodField()
    output = serializers.SerializerMethodField()

    class Meta(object):
        model = NotImplemented
        fields = ('uuid', 'output', 'state', 'created', 'modified', 'request_type',)
        read_only_fields = ('uuid', 'output', 'state', 'created', 'modified', 'request_type',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def get_output(self, obj):
        if self.context.get('select_output'):
            return obj.output
        else:
            return None

    def get_request_type(self, obj):
        return REQUEST_TYPES_PLAIN_NAMES.get(type(obj))

    def get_state(self, obj):
        return obj.human_readable_state


class JupyterHubManagementSyncConfigurationRequestSerializer(JupyterHubManagementRequestMixin):
    class Meta(JupyterHubManagementRequestMixin.Meta):
        model = models.JupyterHubManagementSyncConfigurationRequest


class JupyterHubManagementDeleteRequestSerializer(JupyterHubManagementRequestMixin):
    class Meta(JupyterHubManagementRequestMixin.Meta):
        model = models.JupyterHubManagementDeleteRequest


class JupyterHubManagementMakeVirtualEnvironmentGlobalRequestSerializer(JupyterHubManagementRequestMixin):
    class Meta(JupyterHubManagementRequestMixin.Meta):
        model = models.JupyterHubManagementMakeVirtualEnvironmentGlobalRequest
        fields = JupyterHubManagementRequestMixin.Meta.fields + ('virtual_env_name',)


class JupyterHubManagementMakeVirtualEnvironmentLocalRequestSerializer(JupyterHubManagementRequestMixin):
    class Meta(JupyterHubManagementRequestMixin.Meta):
        model = models.JupyterHubManagementMakeVirtualEnvironmentLocalRequest
        fields = JupyterHubManagementRequestMixin.Meta.fields + ('virtual_env_name',)


class JupyterHubOAuthConfigSerializer(core_serializers.AugmentedSerializerMixin,
                                      structure_serializers.PermissionFieldFilteringMixin,
                                      serializers.HyperlinkedModelSerializer):
    type = serializers.ChoiceField(choices=models.JupyterHubOAuthType.CHOICES)

    class Meta(object):
        model = models.JupyterHubOAuthConfig
        fields = ('uuid', 'type', 'oauth_callback_url', 'client_id', 'client_secret', 'tenant_id', 'gitlab_host',)
        read_only_fields = ('uuid',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def get_filtered_field_names(self):
        return 'type'


class JupyterHubManagementSerializer(
        common_serializers.BaseApplicationSerializer,
        structure_serializers.PermissionFieldFilteringMixin):
    REQUEST_IN_PROGRESS_STATES = (core_models.StateMixin.States.CREATION_SCHEDULED, core_models.StateMixin.States.CREATING)

    python_management = serializers.HyperlinkedRelatedField(
        lookup_field='uuid',
        view_name='python_management-detail',
        queryset=python_management_models.PythonManagement.objects.all(),
    )
    state = serializers.SerializerMethodField()
    jupyter_hub_users = JupyterHubUserSerializer(many=True)
    jupyter_hub_oauth_config = JupyterHubOAuthConfigSerializer(allow_null=True)
    updated_virtual_environments = python_management_serializers.VirtualEnvironmentSerializer(many=True, write_only=True)
    name = serializers.SerializerMethodField()
    jupyter_hub_url = serializers.SerializerMethodField()

    class Meta(object):
        model = models.JupyterHubManagement
        fields = ('uuid', 'python_management', 'jupyter_hub_users', 'state', 'session_time_to_live_hours',
                  'created', 'modified', 'updated_virtual_environments', 'name', 'jupyter_hub_url', 'jupyter_hub_oauth_config',)
        read_only_fields = ('request_states', 'created', 'modified', 'jupyter_hub_url',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def get_filtered_field_names(self):
        return 'project'

    def get_jupyter_hub_url(self, jupyter_hub_management):
        instance_floating_ips = jupyter_hub_management.instance.floating_ips if jupyter_hub_management.instance else None
        return instance_floating_ips[0].address if instance_floating_ips else None

    def get_name(self, jupyter_hub_management):
        instance_name = jupyter_hub_management.instance.name if jupyter_hub_management.instance else 'removed instance'
        return 'JupyterHub - %s - %s' % (jupyter_hub_management.python_management.virtual_envs_dir_path, instance_name)

    def get_state(self, jupyter_hub_management):
        states = []
        configuration_request = pythhon_management_utils.execute_safely(
            lambda: models.JupyterHubManagementSyncConfigurationRequest.objects.filter(jupyter_hub_management=jupyter_hub_management).latest('id'))
        if configuration_request and self.is_in_progress_or_errored(configuration_request):
            return [self.build_state(configuration_request)]

        states.extend(self.get_request_state(
            pythhon_management_utils.execute_safely(
                lambda: models.JupyterHubManagementDeleteRequest.objects.filter(jupyter_hub_management=jupyter_hub_management).latest('id'))))
        states.extend(self.build_states_from_last_group_of_the_request(jupyter_hub_management))

        if not states:
            return core_models.StateMixin(state=core_models.StateMixin.States.OK).human_readable_state
        else:
            creation_scheduled_state = core_models.StateMixin(state=core_models.StateMixin.States.CREATION_SCHEDULED).human_readable_state
            creating_state = core_models.StateMixin(state=core_models.StateMixin.States.CREATING).human_readable_state
            erred_state = core_models.StateMixin(state=core_models.StateMixin.States.ERRED).human_readable_state
            if creating_state in states:
                return creating_state
            elif creation_scheduled_state in states:
                return creation_scheduled_state
            elif erred_state in states:
                return erred_state

    def build_states_from_last_group_of_the_request(self, jupyter_hub_management):
        states = []
        global_requests = models.JupyterHubManagementMakeVirtualEnvironmentGlobalRequest.objects.filter(jupyter_hub_management=jupyter_hub_management).order_by('-pk')
        local_requests = models.JupyterHubManagementMakeVirtualEnvironmentLocalRequest.objects.filter(jupyter_hub_management=jupyter_hub_management).order_by('-pk')
        merged_requests = list(chain(global_requests, local_requests))
        merged_requests.sort(key=lambda r: r.pk, reverse=True)
        last_request_group = self.get_last_requests_group(merged_requests)
        for request in last_request_group:
            if self.is_in_progress_or_errored(request):
                states.append(self.build_state(request))
        return states

    def get_request_state(self, request):
        if request and self.is_in_progress_or_errored(request):
            return [self.build_state(request)]
        else:
            return []

    def get_last_requests_group(self, requests):
        last_request_group = []

        last_request_time = None
        for request in requests:
            if not last_request_time:
                last_request_time = request.created - datetime.timedelta(minutes=1)
            if request.created < last_request_time:
                break
            last_request_group.append(request)

        return last_request_group

    def is_in_progress_or_errored(self, request):
        return request.state in JupyterHubManagementSerializer.REQUEST_IN_PROGRESS_STATES \
            or request.state == core_models.StateMixin.States.ERRED

    def build_state(self, request, state=None):
        request_with_state = state if state else request
        return request_with_state.human_readable_state

    @transaction.atomic
    def create(self, validated_data):
        oauth_config = validated_data.get('jupyter_hub_oauth_config')
        persisted_oauth_config = None
        if not oauth_config:
            for jupyter_hub_user in validated_data.get('jupyter_hub_users'):
                jupyter_hub_user['password'] = sha512_crypt.hash(jupyter_hub_user['password'])
        else:
            for jupyter_hub_user in validated_data.get('jupyter_hub_users'):
                # NB! Should be consistent with JupyterHub configuration located in jupyterhub_config.py.j2
                self.normalize_username(jupyter_hub_user)
            persisted_oauth_config = models.JupyterHubOAuthConfig(
                type=oauth_config['type'],
                oauth_callback_url=oauth_config['oauth_callback_url'],
                client_id=oauth_config['client_id'],
                client_secret=oauth_config['client_secret'],
                tenant_id=oauth_config['tenant_id'] if oauth_config['type'] == models.JupyterHubOAuthType.AZURE else None,
                gitlab_host=oauth_config['gitlab_host'] if oauth_config['type'] == models.JupyterHubOAuthType.GITLAB else None)
            persisted_oauth_config.save()
        jupyter_hub_management = models.JupyterHubManagement(
            user=validated_data.get('user'),
            python_management=validated_data.get('python_management'),
            session_time_to_live_hours=validated_data.get('session_time_to_live_hours'),
            instance=validated_data.get('python_management').instance,
            project=validated_data.get('python_management').instance.service_project_link.project,
            jupyter_hub_oauth_config=persisted_oauth_config)
        jupyter_hub_management.save()

        for user in validated_data.get('jupyter_hub_users'):
            models.JupyterHubUser.objects.create(
                username=user['username'].lower(),
                password=user.get('password', None),
                admin=user['admin'],
                whitelisted=user['whitelisted'],
                jupyter_hub_management=jupyter_hub_management)

        return jupyter_hub_management

    def update(self, instance, validated_data):
        instance.session_time_to_live_hours = validated_data.get('session_time_to_live_hours')
        jupyter_hub_users = validated_data.get('jupyter_hub_users')

        persisted_jupyter_hub_users = instance.jupyter_hub_users.all()
        removed_jupyter_hub_users = jupyter_hub_management_service.JupyterHubManagementService().find_removed_users(persisted_jupyter_hub_users, jupyter_hub_users)
        for removed_jupyter_hub_user in removed_jupyter_hub_users:
            removed_jupyter_hub_user.delete()

        for jupyter_hub_user in jupyter_hub_users:
            self.normalize_username(jupyter_hub_user)
            corresponding_persisted_user = self.find_corresponding_persisted_jupyter_hub_user(persisted_jupyter_hub_users, jupyter_hub_user['username'])
            if corresponding_persisted_user:
                corresponding_user = corresponding_persisted_user[0]
                corresponding_user.username = jupyter_hub_user['username']
                corresponding_user.whitelisted = jupyter_hub_user['whitelisted']
                if jupyter_hub_user['password']:
                    corresponding_user.password = sha512_crypt.hash(jupyter_hub_user['password'])
                corresponding_user.admin = jupyter_hub_user['admin']
                corresponding_user.save()
            else:
                new_jupyter_hub_user = models.JupyterHubUser(
                    jupyter_hub_management=instance,
                    username=jupyter_hub_user['username'],
                    password=sha512_crypt.hash(jupyter_hub_user['password']) if not instance.jupyter_hub_oauth_config else None,
                    admin=jupyter_hub_user['admin'],
                    whitelisted=jupyter_hub_user['whitelisted'])
                new_jupyter_hub_user.save()

        oauth_config = validated_data.get('jupyter_hub_oauth_config')
        if oauth_config:
            instance.jupyter_hub_oauth_config.type = oauth_config['type']
            instance.jupyter_hub_oauth_config.oauth_callback_url = oauth_config['oauth_callback_url']
            instance.jupyter_hub_oauth_config.client_id = oauth_config['client_id']
            instance.jupyter_hub_oauth_config.client_secret = oauth_config['client_secret']
            instance.jupyter_hub_oauth_config.tenant_id = oauth_config['tenant_id'] if oauth_config['type'] == models.JupyterHubOAuthType.AZURE else None
            instance.jupyter_hub_oauth_config.gitlab_host = oauth_config['gitlab_host'] if oauth_config['type'] == models.JupyterHubOAuthType.GITLAB else None
            instance.jupyter_hub_oauth_config.save()

        instance.save()

        return instance

    def normalize_username(self, jupyter_hub_user):
        jupyter_hub_user['username'] = ''.join(c if c.isalnum() else '_' for c in jupyter_hub_user['username'].lower())

    def find_corresponding_persisted_jupyter_hub_user(self, persisted_jupyter_hub_users, username):
        return filter(lambda u: u.username == username.lower(), persisted_jupyter_hub_users)

    def validate(self, attrs):
        super(JupyterHubManagementSerializer, self).validate(attrs)
        if not self.instance:
            attrs['user'] = self.context['request'].user

        self.check_project_permissions(attrs)
        return attrs

    def check_project_permissions(self, attrs):
        if self.instance:
            project = self.instance.project
        else:
            project = attrs['python_management'].project

        if not structure_permissions._has_admin_access(self.context['request'].user, project):
            raise exceptions.PermissionDenied()


class SummaryJupyterHubManagementRequestsSerializer(core_serializers.BaseSummarySerializer):
    @classmethod
    def get_serializer(cls, model):
        if model is models.JupyterHubManagementSyncConfigurationRequest:
            return JupyterHubManagementSyncConfigurationRequestSerializer
        elif model is models.JupyterHubManagementMakeVirtualEnvironmentGlobalRequest:
            return JupyterHubManagementMakeVirtualEnvironmentGlobalRequestSerializer
        elif model is models.JupyterHubManagementMakeVirtualEnvironmentLocalRequest:
            return JupyterHubManagementMakeVirtualEnvironmentLocalRequestSerializer
        elif model is models.JupyterHubManagementDeleteRequest:
            return JupyterHubManagementDeleteRequestSerializer



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-10-23 11:27
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0027_non_unique_template_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='IgnoredIssueStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, unique=True, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
            ],
        ),
        migrations.CreateModel(
            name='TemplateStatusNotification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.CharField(max_length=255, unique=True, validators=[waldur_core.core.validators.validate_name])),
                ('html', models.TextField(validators=[waldur_core.core.validators.validate_name])),
                ('text', models.TextField(validators=[waldur_core.core.validators.validate_name])),
                ('subject', models.CharField(max_length=255, validators=[waldur_core.core.validators.validate_name])),
            ],
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


def migrate_invoice_item_price(apps, schema_editor):
    OpenStackItem = apps.get_model('invoices', 'OpenStackItem')
    OfferingItem = apps.get_model('invoices', 'OfferingItem')

    OpenStackItem.objects.all().update(unit_price=models.F('daily_price'))
    OfferingItem.objects.all().update(unit_price=models.F('daily_price'))


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0015_add_unit_price'),
        ('support', '0011_remove_price'),
    ]

    operations = [
        migrations.RunPython(migrate_invoice_item_price, elidable=True),
        migrations.RemoveField(
            model_name='offeringitem',
            name='daily_price',
        ),
        migrations.RemoveField(
            model_name='openstackitem',
            name='daily_price',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-02-12 13:48
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_slurm', '0004_increase_precision'),
    ]

    operations = [
        migrations.AddField(
            model_name='allocation',
            name='deposit_limit',
            field=models.DecimalField(decimal_places=0, default=-1, max_digits=6),
        ),
        migrations.AddField(
            model_name='allocation',
            name='deposit_usage',
            field=models.DecimalField(decimal_places=2, default=0, max_digits=8),
        ),
    ]



import logging

from django.core.exceptions import ObjectDoesNotExist

from waldur_azure import models as azure_models
from waldur_core.core import utils as core_utils

from . import utils


logger = logging.getLogger(__name__)


def synchronize_nic(sender, instance, created=False, **kwargs):
    nic = instance
    if not created and not set(nic.tracker.changed()) & {'public_ip_id', 'ip_address'}:
        return

    utils.synchronize_nic(nic)


def synchronize_public_ip(sender, instance, created=False, **kwargs):
    public_ip = instance

    if not created and not public_ip.tracker.has_changed('ip_address'):
        return

    try:
        nic = azure_models.NetworkInterface.objects.get(public_ip=public_ip)
        utils.synchronize_nic(nic)
    except ObjectDoesNotExist:
        logger.debug('Skipping resource synchronization for Azure virtual machine'
                     'because marketplace resource does not exist. '
                     'Resource: %s', core_utils.serialize_instance(public_ip))
        return



from django.apps import AppConfig
from django.db.models import signals


class PayPalConfig(AppConfig):
    name = 'waldur_paypal'
    verbose_name = 'PayPal'

    def ready(self):
        from . import handlers
        from waldur_core.cost_tracking import signals as cost_signals
        from django.conf import settings

        if not settings.WALDUR_PAYPAL['ENABLED']:
            return

        Invoice = self.get_model('Invoice')

        signals.post_save.connect(
            handlers.log_invoice_save,
            sender=Invoice,
            dispatch_uid='waldur_paypal.handlers.log_invoice_save',
        )

        signals.post_delete.connect(
            handlers.log_invoice_delete,
            sender=Invoice,
            dispatch_uid='waldur_paypal.handlers.log_invoice_delete',
        )

        cost_signals.invoice_created.connect(
            handlers.create_invoice,
            sender=None,
            dispatch_uid='waldur_paypal.handlers.create_invoice',
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-08-09 10:34
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0022_offering_template'),
    ]

    def migrate_config(apps, schema_editor):
        if settings.DATABASES['default']['ENGINE'] != 'django.db.backends.postgresql_psycopg2':
            return 

        OfferingTemplate = apps.get_model('support', 'OfferingTemplate')
        Offering = apps.get_model('support', 'Offering')
        config = settings.WALDUR_SUPPORT['OFFERINGS']

        for offering in Offering.objects.all():
            try:
                template, created = OfferingTemplate.objects.get_or_create(name=offering.type,
                                                                           defaults={'config': config[offering.type]})
                offering.template = template
                offering.save()
            except KeyError:
                pass

        for template in config.keys():
            OfferingTemplate.objects.get_or_create(name=template,
                                                   defaults={'config': config[template]})

    operations = [
        migrations.RunPython(migrate_config, elidable=True),
        migrations.AlterField(
            model_name='offering',
            name='template',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='support.OfferingTemplate')
        ),
    ]



import logging

from django.conf import settings
from waldur_ansible.common import backend as common_backend
from waldur_ansible.python_management import models, constants, executors

from . import output_lines_post_processors, locking_service, extracted_information_handlers, additional_extra_args_builders, error_handlers

logger = logging.getLogger(__name__)


class PythonManagementBackend(common_backend.ManagementRequestsBackend):
    REQUEST_TYPES_PLAYBOOKS_CORRESPONDENCE = {
        models.PythonManagementInitializeRequest: constants.PythonManagementConstants.INSTALL_PYTHON_ENVIRONMENT,
        models.PythonManagementSynchronizeRequest: constants.PythonManagementConstants.SYNCHRONIZE_PACKAGES,
        models.PythonManagementFindVirtualEnvsRequest: constants.PythonManagementConstants.FIND_INSTALLED_VIRTUAL_ENVIRONMENTS,
        models.PythonManagementFindInstalledLibrariesRequest: constants.PythonManagementConstants.FIND_INSTALLED_LIBRARIES_FOR_VIRTUAL_ENVIRONMENT,
        models.PythonManagementDeleteVirtualEnvRequest: constants.PythonManagementConstants.DELETE_VIRTUAL_ENVIRONMENT,
        models.PythonManagementDeleteRequest: constants.PythonManagementConstants.DELETE_PYTHON_ENVIRONMENT,
    }

    REQUEST_TYPES_EXTRA_ARGS_CORRESPONDENCE = {
        models.PythonManagementInitializeRequest: None,
        models.PythonManagementSynchronizeRequest: additional_extra_args_builders.build_sync_request_extra_args,
        models.PythonManagementFindVirtualEnvsRequest: None,
        models.PythonManagementFindInstalledLibrariesRequest: additional_extra_args_builders.build_additional_extra_args,
        models.PythonManagementDeleteVirtualEnvRequest: additional_extra_args_builders.build_additional_extra_args,
        models.PythonManagementDeleteRequest: None,
    }

    REQUEST_TYPES_POST_PROCESSOR_CORRESPONDENCE = {
        models.PythonManagementInitializeRequest: output_lines_post_processors.InitializationOutputLinesPostProcessor,
        models.PythonManagementSynchronizeRequest: output_lines_post_processors.InstalledLibrariesOutputLinesPostProcessor,
        models.PythonManagementFindVirtualEnvsRequest: output_lines_post_processors.InstalledVirtualEnvironmentsOutputLinesPostProcessor,
        models.PythonManagementFindInstalledLibrariesRequest: output_lines_post_processors.InstalledLibrariesOutputLinesPostProcessor,
        models.PythonManagementDeleteVirtualEnvRequest: output_lines_post_processors.NullOutputLinesPostProcessor,
        models.PythonManagementDeleteRequest: output_lines_post_processors.NullOutputLinesPostProcessor,
    }

    REQUEST_TYPES_HANDLERS_CORRESPONDENCE = {
        models.PythonManagementInitializeRequest: extracted_information_handlers.InitializationRequestExtractedInformationHandler,
        models.PythonManagementSynchronizeRequest: extracted_information_handlers.InstalledLibrariesExtractedInformationHandler,
        models.PythonManagementFindVirtualEnvsRequest: extracted_information_handlers.PythonManagementFindVirtualEnvsRequestExtractedInformationHandler,
        models.PythonManagementFindInstalledLibrariesRequest: extracted_information_handlers.InstalledLibrariesExtractedInformationHandler,
        models.PythonManagementDeleteVirtualEnvRequest: extracted_information_handlers.PythonManagementDeleteVirtualEnvExtractedInformationHandler,
        models.PythonManagementDeleteRequest: extracted_information_handlers.PythonManagementDeletionRequestExtractedInformationHandler,
    }

    REQUEST_TYPES_ERROR_HANDLERS_CORRESPONDENCE = {
        models.PythonManagementInitializeRequest: error_handlers.NullErrorHandler,
        models.PythonManagementSynchronizeRequest: error_handlers.NullErrorHandler,
        models.PythonManagementFindVirtualEnvsRequest: error_handlers.NullErrorHandler,
        models.PythonManagementFindInstalledLibrariesRequest: error_handlers.NullErrorHandler,
        models.PythonManagementDeleteVirtualEnvRequest: error_handlers.NullErrorHandler,
        models.PythonManagementDeleteRequest: error_handlers.DeleteRequestErrorHandler,
    }

    LOCKED_FOR_PROCESSING = 'Whole environment or the particular virutal environnment is now being processed, request cannot be executed!'

    def process_python_management_request(self, python_management_request):
        self.process_request(python_management_request)

    def is_processing_allowed(self, request):
        return locking_service.PythonManagementBackendLockingService.is_processing_allowed(request)

    def build_locked_for_processing_message(self, request):
        return PythonManagementBackend.LOCKED_FOR_PROCESSING

    def lock_for_processing(self, request):
        locking_service.PythonManagementBackendLockingService.lock_for_processing(request)

    def handle_on_processing_finished(self, request):
        locking_service.PythonManagementBackendLockingService.handle_on_processing_finished(request)

    def get_playbook_path(self, request):
        return settings.WALDUR_PYTHON_MANAGEMENT.get('PYTHON_MANAGEMENT_PLAYBOOKS_DIRECTORY') \
            + PythonManagementBackend.REQUEST_TYPES_PLAYBOOKS_CORRESPONDENCE.get(type(request)) \
            + '.yml'

    def get_user(self, request):
        return request.python_management.user

    def build_additional_extra_vars(self, request):
        python_management = request.python_management
        extra_vars = dict(
            instance_public_ip=python_management.instance.external_ips[0] if python_management.instance else None,
            virtual_envs_dir_path=python_management.virtual_envs_dir_path,
            default_system_user=python_management.system_user,
        )

        additional_extra_args_building_function = PythonManagementBackend.REQUEST_TYPES_EXTRA_ARGS_CORRESPONDENCE.get(type(request))
        if additional_extra_args_building_function:
            extra_vars.update(additional_extra_args_building_function(request))

        return extra_vars

    def instantiate_line_post_processor_class(self, request):
        lines_post_processor_class = PythonManagementBackend.REQUEST_TYPES_POST_PROCESSOR_CORRESPONDENCE \
            .get(type(request))
        return lines_post_processor_class()

    def instantiate_extracted_information_handler_class(self, request):
        extracted_information_handler_class = PythonManagementBackend.REQUEST_TYPES_HANDLERS_CORRESPONDENCE \
            .get(type(request))
        return extracted_information_handler_class()

    def instantiate_error_handler_class(self, request):
        extracted_information_handler_class = PythonManagementBackend.REQUEST_TYPES_ERROR_HANDLERS_CORRESPONDENCE.get(type(request))
        return extracted_information_handler_class()


class PythonManagementInitializationBackend(PythonManagementBackend):

    def process_python_management_request(self, python_management_initialization_request):
        super(PythonManagementInitializationBackend, self).process_python_management_request(python_management_initialization_request)

        for synchronization_request in python_management_initialization_request.sychronization_requests.all():
            executors.PythonManagementRequestExecutor.execute(synchronization_request, async=True)



import django_filters

from waldur_core.core import filters as core_filters

from . import models


class DeploymentPlanFilter(django_filters.FilterSet):
    project = core_filters.URLFilter(view_name='project-detail', name='project__uuid')
    project_uuid = django_filters.UUIDFilter(name='project__uuid')
    customer = core_filters.URLFilter(view_name='customer-detail', name='project__customer__uuid')
    customer_uuid = django_filters.UUIDFilter(name='project__customer__uuid')

    o = django_filters.OrderingFilter(
        fields=('name', 'created')
    )

    class Meta(object):
        model = models.DeploymentPlan
        fields = ('project', 'project_uuid', 'customer', 'customer_uuid')



from waldur_core.core import executors as core_executors
from waldur_core.core import tasks as core_tasks


class ClusterCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_instance,
            'create_cluster',
            state_transition='begin_creating'
        )


class ClusterDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if instance.backend_id:
            return core_tasks.BackendMethodTask().si(
                serialized_instance,
                'delete_cluster',
                state_transition='begin_deleting')
        else:
            return core_tasks.StateTransitionTask().si(
                serialized_instance,
                state_transition='begin_deleting'
            )


class ClusterUpdateExecutor(core_executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if instance.backend_id and {'name'} & set(kwargs['updated_fields']):
            return core_tasks.BackendMethodTask().si(
                serialized_instance,
                'update_cluster',
                state_transition='begin_updating')
        else:
            return core_tasks.StateTransitionTask().si(
                serialized_instance,
                state_transition='begin_updating'
            )



from __future__ import unicode_literals

from celery import chain

from waldur_core.core import executors as core_executors, \
    tasks as core_tasks, utils as core_utils
from waldur_core.structure import executors as structure_executors

from . import models


class VirtualMachineCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        serialized_resource_group = core_utils.serialize_instance(instance.resource_group)
        serialized_storage_account = core_utils.serialize_instance(
            instance.resource_group.storageaccount_set.get())
        serialized_network = core_utils.serialize_instance(
            instance.network_interface.subnet.network)
        serialized_subnet = core_utils.serialize_instance(
            instance.network_interface.subnet)
        serialized_public_ip = core_utils.serialize_instance(
            instance.network_interface.public_ip)
        serialized_nic = core_utils.serialize_instance(
            instance.network_interface)

        return chain(
            core_tasks.StateTransitionTask().si(
                serialized_instance,
                state_transition='begin_creating'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_resource_group,
                backend_method='create_resource_group',
                state_transition='begin_creating',
            ),
            core_tasks.StateTransitionTask().si(
                serialized_resource_group,
                state_transition='set_ok',
            ),
            core_tasks.BackendMethodTask().si(
                serialized_storage_account,
                backend_method='create_storage_account',
                state_transition='begin_creating',
            ),
            core_tasks.StateTransitionTask().si(
                serialized_storage_account,
                state_transition='set_ok',
            ),
            core_tasks.BackendMethodTask().si(
                serialized_network,
                backend_method='create_network',
                state_transition='begin_creating',
            ),
            core_tasks.StateTransitionTask().si(
                serialized_network,
                state_transition='set_ok',
            ),
            core_tasks.BackendMethodTask().si(
                serialized_subnet,
                backend_method='create_subnet',
                state_transition='begin_creating',
            ),
            core_tasks.StateTransitionTask().si(
                serialized_subnet,
                state_transition='set_ok',
            ),
            core_tasks.BackendMethodTask().si(
                serialized_public_ip,
                backend_method='create_public_ip',
                state_transition='begin_creating',
            ),
            core_tasks.StateTransitionTask().si(
                serialized_public_ip,
                state_transition='set_ok',
            ),
            core_tasks.BackendMethodTask().si(
                serialized_nic,
                backend_method='create_network_interface',
                state_transition='begin_creating',
            ),
            core_tasks.StateTransitionTask().si(
                serialized_nic,
                state_transition='set_ok',
            ),
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='create_virtual_machine',
            ),
            core_tasks.BackendMethodTask().si(
                serialized_nic,
                backend_method='pull_network_interface',
            ),
            core_tasks.BackendMethodTask().si(
                serialized_public_ip,
                backend_method='pull_public_ip_address',
            ),
        )


class VirtualMachineDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if instance.backend_id:
            return core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='delete_virtual_machine',
                state_transition='begin_deleting',
            )
        else:
            return core_tasks.StateTransitionTask().si(
                serialized_instance, state_transition='begin_deleting')


class VirtualMachineStartExecutor(core_executors.ActionExecutor):
    action = 'Start'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_instance,
            backend_method='start_virtual_machine',
            state_transition='begin_updating',
        )


class VirtualMachineStopExecutor(core_executors.ActionExecutor):
    action = 'Stop'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_instance,
            backend_method='stop_virtual_machine',
            state_transition='begin_updating',
        )


class VirtualMachineRestartExecutor(core_executors.ActionExecutor):
    action = 'Restart'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_instance,
            backend_method='restart_virtual_machine',
            state_transition='begin_updating',
        )


class PublicIPCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_instance,
            backend_method='create_public_ip',
            state_transition='begin_creating',
        )


class PublicIPDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if instance.backend_id:
            return core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='delete_public_ip',
                state_transition='begin_deleting',
            )
        else:
            return core_tasks.StateTransitionTask().si(
                serialized_instance, state_transition='begin_deleting')


class SQLServerCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        serialized_resource_group = core_utils.serialize_instance(instance.resource_group)
        return chain(
            core_tasks.StateTransitionTask().si(
                serialized_instance,
                state_transition='begin_creating'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_resource_group,
                backend_method='create_resource_group',
                state_transition='begin_creating',
            ),
            core_tasks.StateTransitionTask().si(
                serialized_resource_group,
                state_transition='set_ok',
            ),
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='create_pgsql_server',
            ),
        )


class SQLServerDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if instance.backend_id:
            return core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='delete_pgsql_server',
                state_transition='begin_deleting',
            )
        else:
            return core_tasks.StateTransitionTask().si(
                serialized_instance, state_transition='begin_deleting')


class SQLDatabaseCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_instance,
            backend_method='create_pgsql_database',
            state_transition='begin_creating',
        )


class SQLDatabaseDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if instance.backend_id:
            return core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='delete_pgsql_database',
                state_transition='begin_deleting',
            )
        else:
            return core_tasks.StateTransitionTask().si(
                serialized_instance, state_transition='begin_deleting')


class AzureCleanupExecutor(structure_executors.BaseCleanupExecutor):
    executors = (
        (models.VirtualMachine, VirtualMachineDeleteExecutor),
        (models.SQLDatabase, SQLDatabaseDeleteExecutor),
        (models.SQLServer, SQLServerDeleteExecutor),
    )



from django.db import models

from waldur_core.core import fields as core_fields


class IdentityProvider(models.Model):
    name = models.TextField(db_index=True)
    url = models.URLField()
    metadata = core_fields.JSONField(default=dict)

    class Meta(object):
        ordering = ('name',)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-09-09 13:11
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


def import_quotas(apps, schema_editor):
    from waldur_mastermind.analytics.utils import import_daily_usage

    import_daily_usage()


class Migration(migrations.Migration):

    replaces = [(b'analytics', '0001_initial'), (b'analytics', '0002_import_quotas'), (b'analytics', '0003_daily_quota_constraints')]

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('quotas', '0001_squashed_0004'),
        ('reversion', '0001_squashed_0004_auto_20160611_1202'),
    ]

    operations = [
        migrations.CreateModel(
            name='DailyQuotaHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField(null=True)),
                ('name', models.CharField(db_index=True, max_length=150)),
                ('usage', models.BigIntegerField()),
                ('date', models.DateField()),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.RunPython(import_quotas),
        migrations.AlterField(
            model_name='dailyquotahistory',
            name='content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType'),
        ),
        migrations.AlterField(
            model_name='dailyquotahistory',
            name='object_id',
            field=models.PositiveIntegerField(),
        ),
        migrations.AlterUniqueTogether(
            name='dailyquotahistory',
            unique_together=set([('content_type', 'object_id', 'name', 'date')]),
        ),
    ]



import logging

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from django.template import Context, Template

from waldur_mastermind.marketplace import models as marketplace_models
from waldur_mastermind.marketplace_support import PLUGIN_NAME
from waldur_mastermind.support import models as support_models


logger = logging.getLogger(__name__)


def get_match_states():
    return {
        support_models.Offering.States.REQUESTED: marketplace_models.Resource.States.CREATING,
        support_models.Offering.States.OK: marketplace_models.Resource.States.OK,
        support_models.Offering.States.TERMINATED: marketplace_models.Resource.States.TERMINATED,
    }


def init_offerings_and_resources(category, customer):
    offerings_counter = 0
    plans_counter = 0
    resources_counter = 0

    # Import marketplace offerings
    ct = ContentType.objects.get_for_model(support_models.OfferingTemplate)
    exist_ids = marketplace_models.Offering.objects.filter(content_type=ct).values_list('object_id', flat=True)

    for template in support_models.OfferingTemplate.objects.exclude(id__in=exist_ids):
        marketplace_models.Offering.objects.create(
            scope=template,
            type=PLUGIN_NAME,
            name=template.config.get('label', template.name),
            customer=customer,
            category=category,
            state=marketplace_models.Offering.States.ACTIVE,
            options={
                'order': template.config.get('order', []),
                'options': template.config.get('options', {})
            },
            description=template.config.get('description', ''),
            full_description=template.config.get('summary', ''),
            terms_of_service=template.config.get('terms_of_service', ''),
        )
        offerings_counter += 1

    # Import marketplace resources
    ct = ContentType.objects.get_for_model(support_models.Offering)
    exist_ids = marketplace_models.Resource.objects.filter(content_type=ct).values_list('object_id', flat=True)

    for support_offering in support_models.Offering.objects.exclude(id__in=exist_ids):
        # get offering
        offering = marketplace_models.Offering.objects.get(scope=support_offering.template, type=PLUGIN_NAME)

        # get plan
        for offering_plan in support_offering.template.plans.all():
            new_plan, create = marketplace_models.Plan.objects.get_or_create(
                scope=offering_plan,
                defaults=dict(
                    offering=offering,
                    name=offering_plan.name,
                    unit_price=offering_plan.unit_price,
                    unit=offering_plan.unit,
                    product_code=offering_plan.product_code,
                    article_code=offering_plan.article_code
                )
            )
            if create:
                plans_counter += 1

        try:
            marketplace_plan = offering.plans.get(unit_price=support_offering.unit_price, unit=support_offering.unit)
        except marketplace_models.Plan.DoesNotExist:
            marketplace_plan = marketplace_models.Plan.objects.create(
                scope=support_offering,
                offering=offering,
                name=support_offering.name,
                unit_price=support_offering.unit_price,
                unit=support_offering.unit,
                product_code=support_offering.product_code,
                article_code=support_offering.article_code,
            )
            plans_counter += 1

        # create resource
        resource = marketplace_models.Resource.objects.create(
            name=support_offering.name,
            created=support_offering.created,
            project=support_offering.project,
            offering=offering,
            plan=marketplace_plan,
            scope=support_offering,
            state=get_match_states()[support_offering.state],
            attributes={'summary': support_offering.issue.summary,
                        'description': support_offering.issue.description,
                        'name': support_offering.name}
        )
        if marketplace_plan:
            marketplace_models.ResourcePlanPeriod.objects.create(
                resource=resource,
                plan=marketplace_plan,
                start=support_offering.created,
                end=support_offering.terminated_at,
            )
        resources_counter += 1

    return offerings_counter, plans_counter, resources_counter


def get_issue_resource(issue):
    """
    This utility method is called from support backend create_issue method
    to traverse from support issue to marketplace resource.

    1) When marketplace resource is created, we need to fetch
    related support offering and then marketplace resource:

    Support Issue <- Support Offering <- Marketplace Resource

    2) When marketplace resource plan is switched, we need to traverse
    from support issue to marketplace order item and then marketplace resource:

    Support issue -> Marketplace Order Item -> Marketplace resource
    """

    if isinstance(issue.resource, marketplace_models.OrderItem):
        return issue.resource.resource
    try:
        offering = support_models.Offering.objects.get(issue=issue)
        return marketplace_models.Resource.objects.get(scope=offering)
    except (ObjectDoesNotExist, MultipleObjectsReturned):
        logger.debug('Resource for issue is not found. Issue ID: %s', issue.id)
        return None


def get_request_link(request):
    link_template = settings.WALDUR_MARKETPLACE_SUPPORT['REQUEST_LINK_TEMPLATE']
    return link_template.format(request_uuid=request.uuid)


def format_description(template_name, context):
    template = Template(settings.WALDUR_MARKETPLACE_SUPPORT[template_name])
    return template.render(Context(context, autoescape=False))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-08-27 09:42
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_vmware', '0020_guest_power_state'),
    ]

    operations = [
        migrations.AddField(
            model_name='virtualmachine',
            name='tools_state',
            field=models.CharField(blank=True, choices=[('STARTING', 'Starting'), ('RUNNING', 'Running'), ('NOT_RUNNING', 'Not running')], max_length=50, verbose_name='Current running status of VMware Tools running in the guest operating system.'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-04-12 15:42
from __future__ import unicode_literals

from decimal import Decimal
import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0077_plan_period_uuid_unique'),
    ]

    operations = [
        migrations.AlterField(
            model_name='plancomponent',
            name='price',
            field=models.DecimalField(decimal_places=7, default=0, max_digits=15, validators=[django.core.validators.MinValueValidator(Decimal('0'))], verbose_name='Price per unit per billing period.'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-06-19 08:26
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.validators
from waldur_core.media.models import get_upload_path


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0006_allow_svg_category_icon'),
    ]

    operations = [
        migrations.CreateModel(
            name='Screenshots',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('image', models.ImageField(upload_to=get_upload_path)),
                ('thumbnail', models.ImageField(editable=False, null=True, upload_to=get_upload_path))
            ],
            options={
                'verbose_name': 'Screenshot',
            },
        ),
        migrations.AddField(
            model_name='screenshots',
            name='offering',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='screenshots', to='marketplace.Offering'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-04 12:14
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0054_categorycomponentusage'),
    ]

    operations = [
        migrations.CreateModel(
            name='OfferingFile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('file', models.FileField(upload_to='offering_files')),
                ('offering', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='files', to='marketplace.Offering')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0002_enlarge_civil_number_user_field'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='registration_method',
            field=models.CharField(default='default', help_text='Indicates what registration method were used.', max_length=50, verbose_name='registration method', blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-09-06 11:33
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0031_plan_component'),
    ]

    operations = [
        migrations.AddField(
            model_name='orderitem',
            name='error_message',
            field=models.TextField(blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('packages', '0004_packagetemplate_category'),
    ]

    operations = [
        migrations.AlterField(
            model_name='openstackpackage',
            name='service_settings',
            field=models.ForeignKey(related_name='+', on_delete=django.db.models.deletion.SET_NULL, to='structure.ServiceSettings', null=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django_fsm


def migrate_instance_state(apps, schema_editor):
    Instance = apps.get_model('openstack', 'Instance')
    # old states
    PROVISIONING_SCHEDULED = 1
    PROVISIONING = 2
    ONLINE = 3
    OFFLINE = 4
    STARTING_SCHEDULED = 5
    STARTING = 6
    STOPPING_SCHEDULED = 7
    STOPPING = 8
    ERRED = 9
    DELETION_SCHEDULED = 10
    DELETING = 11
    RESIZING_SCHEDULED = 13
    RESIZING = 14
    RESTARTING_SCHEDULED = 15
    RESTARTING = 16
    # new states
    NEW_CREATION_SCHEDULED = 5
    NEW_CREATING = 6
    NEW_UPDATE_SCHEDULED = 1
    NEW_UPDATING = 2
    NEW_DELETION_SCHEDULED = 7
    NEW_DELETING = 8
    NEW_OK = 3
    NEW_ERRED = 4

    state_map = {
        PROVISIONING_SCHEDULED: NEW_CREATION_SCHEDULED,
        PROVISIONING: NEW_CREATING,
        ONLINE: NEW_OK,
        OFFLINE: NEW_OK,
        STARTING_SCHEDULED: NEW_UPDATE_SCHEDULED,
        STARTING: NEW_UPDATING,
        STOPPING_SCHEDULED: NEW_UPDATE_SCHEDULED,
        STOPPING: NEW_UPDATING,
        ERRED: NEW_ERRED,
        DELETION_SCHEDULED: NEW_DELETION_SCHEDULED,
        DELETING: NEW_DELETING,
        RESIZING_SCHEDULED: NEW_UPDATE_SCHEDULED,
        RESIZING: NEW_UPDATING,
        RESTARTING_SCHEDULED: NEW_UPDATE_SCHEDULED,
        RESTARTING: NEW_UPDATING,
    }

    for instance in Instance.objects.all():
        instance.state = state_map[instance.state]
        instance.save()


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0022_volume_device'),
    ]

    operations = [
        migrations.RunPython(migrate_instance_state, elidable=True),
        migrations.AlterField(
            model_name='instance',
            name='state',
            field=django_fsm.FSMIntegerField(default=5, choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')]),
        ),
    ]



from waldur_ansible.common.extension import AnsibleCommonExtension

from waldur_core.core import WaldurExtension


class PythonManagementExtension(WaldurExtension):
    class Settings:
        WALDUR_PYTHON_MANAGEMENT = {
            'PYTHON_MANAGEMENT_PLAYBOOKS_DIRECTORY': '%swaldur-apps/python_management/' % AnsibleCommonExtension.Settings.WALDUR_ANSIBLE_COMMON['ANSIBLE_LIBRARY'],
            'SYNC_PIP_PACKAGES_TASK_ENABLED': False,
            'SYNC_PIP_PACKAGES_BATCH_SIZE': 300,
        }

    @staticmethod
    def django_app():
        return 'waldur_ansible.python_management'

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def is_assembly():
        return True

    @staticmethod
    def celery_tasks():
        from datetime import timedelta
        return {
            'waldur-ansible-sync-pip-packages': {
                'task': 'waldur_ansible.sync_pip_libraries',
                'schedule': timedelta(hours=48),
                'args': (),
            },
        }



from . import views


def register_in(router):
    router.register(r'deployment-plans', views.DeploymentPlanViewSet, base_name='deployment-plan')
    router.register(r'deployment-presets', views.PresetViewSet, base_name='deployment-preset')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cost_tracking', '0024_refactor_price_estimate'),
    ]

    operations = [
        migrations.AlterField(
            model_name='defaultpricelistitem',
            name='value',
            field=models.DecimalField(default=0, verbose_name='Hourly rate', max_digits=13, decimal_places=7),
        ),
        migrations.AlterField(
            model_name='pricelistitem',
            name='value',
            field=models.DecimalField(default=0, verbose_name='Hourly rate', max_digits=13, decimal_places=7),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0026_invoice__file'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = []



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models

import waldur_core.core.fields


class Migration(migrations.Migration):
    replaces = [('quotas', '0001_initial'),
                ('quotas', '0002_make_quota_scope_nullable'),
                ('quotas', '0003_index_quota_name'),
                ('quotas', '0004_quota_threshold')]

    initial = True

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Quota',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('threshold', models.FloatField(default=0, validators=[django.core.validators.MinValueValidator(0)])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('limit', models.FloatField(default=-1)),
                ('usage', models.FloatField(default=0)),
                ('name', models.CharField(db_index=True, max_length=150)),
                ('object_id', models.PositiveIntegerField(null=True)),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
            bases=(waldur_core.logging.loggers.LoggableMixin, waldur_core.core.models.ReversionMixin, models.Model),
        ),
        migrations.AlterUniqueTogether(
            name='quota',
            unique_together=set([('name', 'content_type', 'object_id')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0003_paymentdetails_default_tax_percent'),
    ]

    operations = [
        migrations.AddField(
            model_name='paymentdetails',
            name='uuid',
            field=waldur_core.core.fields.UUIDField(),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-03-28 14:32
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0072_offering_component_product_code'),
    ]

    operations = [
        migrations.AddField(
            model_name='componentusage',
            name='plan_period',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, to='marketplace.ResourcePlanPeriod', related_name='components', null=True),
            preserve_default=False,
        ),
        migrations.AlterUniqueTogether(
            name='componentusage',
            unique_together=set([('resource', 'component', 'plan_period')]),
        ),
    ]



from django.dispatch import Signal

# Role related signals
# sender = structure class, e.g. Customer or Project
structure_role_granted = Signal(providing_args=['structure', 'user', 'role', 'created_by'])
structure_role_revoked = Signal(providing_args=['structure', 'user', 'role', 'removed_by'])
structure_role_updated = Signal(providing_args=['instance', 'user'])

resource_imported = Signal(providing_args=['instance', 'plan', 'offering'])



import json
import logging
import re

from ceilometerclient import exc as ceilometer_exceptions
from cinderclient import exceptions as cinder_exceptions
from cinderclient.v2.contrib import list_extensions
from django.db import transaction, IntegrityError
from django.utils import timezone, dateparse
from django.utils.functional import cached_property
from keystoneclient import exceptions as keystone_exceptions
from neutronclient.client import exceptions as neutron_exceptions
from novaclient import exceptions as nova_exceptions

from waldur_core.structure import log_backend_action
from waldur_core.structure.utils import (
    update_pulled_fields, handle_resource_not_found, handle_resource_update_success)
from waldur_openstack.openstack_base.backend import BaseOpenStackBackend, OpenStackBackendError, reraise

from . import models

logger = logging.getLogger(__name__)


def backend_internal_ip_to_internal_ip(backend_internal_ip, **kwargs):
    internal_ip = models.InternalIP(
        backend_id=backend_internal_ip['id'],
        mac_address=backend_internal_ip['mac_address'],
        ip4_address=backend_internal_ip['fixed_ips'][0]['ip_address'],
    )

    for field, value in kwargs.items():
        setattr(internal_ip, field, value)

    if 'instance' not in kwargs:
        internal_ip._instance_backend_id = backend_internal_ip['device_id']
    if 'subnet' not in kwargs:
        internal_ip._subnet_backend_id = backend_internal_ip['fixed_ips'][0]['subnet_id']

    internal_ip._device_owner = backend_internal_ip['device_owner']

    return internal_ip


class InternalIPSynchronizer(object):
    """
    It is assumed that all subnets for the current tenant have been successfully synchronized.
    """

    def __init__(self, neutron_client, tenant_id, settings):
        self.neutron_client = neutron_client
        self.tenant_id = tenant_id
        self.settings = settings

    @cached_property
    def remote_ips(self):
        """
        Fetch all Neutron ports for the current tenant.
        Convert Neutron port to local internal IP model.
        """
        try:
            ips = self.neutron_client.list_ports(tenant_id=self.tenant_id)['ports']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        return [backend_internal_ip_to_internal_ip(ip) for ip in ips]

    @cached_property
    def local_ips(self):
        """
        Prepare mapping from backend ID to local internal IP model.
        """
        internal_ips = models.InternalIP.objects\
            .filter(subnet__settings=self.settings).exclude(backend_id=None)
        return {ip.backend_id: ip for ip in internal_ips}

    @cached_property
    def pending_ips(self):
        """
        Prepare mapping from device and subnet ID to local internal IP model.
        """
        pending_internal_ips = models.InternalIP.objects.filter(
            subnet__settings=self.settings, backend_id=None).exclude(instance__isnull=True)
        return {(ip.instance.backend_id, ip.subnet.backend_id): ip for ip in pending_internal_ips}

    @cached_property
    def stale_ips(self):
        """
        Prepare list of IDs for stale internal IPs.
        """
        remote_ips = {ip.backend_id for ip in self.remote_ips}
        return [
            ip.pk
            for (backend_id, ip) in self.local_ips.items()
            if backend_id not in remote_ips
        ]

    @cached_property
    def instances(self):
        """
        Prepare mapping from backend ID to local instance model.
        """
        instances = models.Instance.objects.filter(
            service_project_link__service__settings=self.settings).exclude(backend_id='')
        return {instance.backend_id: instance for instance in instances}

    @cached_property
    def subnets(self):
        """
        Prepare mapping from backend ID to local subnet model.
        """
        subnets = models.SubNet.objects.filter(settings=self.settings).exclude(backend_id='')
        return {subnet.backend_id: subnet for subnet in subnets}

    @transaction.atomic
    def execute(self):
        for remote_ip in self.remote_ips:

            # Check if related subnet exists.
            subnet = self.subnets.get(remote_ip._subnet_backend_id)
            if subnet is None:
                logger.warning('Skipping Neutron port synchronization process because '
                               'related subnet is not imported yet. Port ID: %s, subnet ID: %s',
                               remote_ip.backend_id, remote_ip._subnet_backend_id)
                continue

            local_ip = self.local_ips.get(remote_ip.backend_id)
            instance = None

            # Please note that for different availability zones you may see not
            # only compute:nova but also, for example, compute:MS-ZONE and compute:RHEL-ZONE
            # therefore it's safer to check prefix only. See also Neutron source code:
            # https://github.com/openstack/neutron/blob/2d30981289474c3e08ff1008b76284a993a57e1f/neutron/plugins/ml2/plugin.py#L2012
            if remote_ip._device_owner.startswith('compute:'):
                instance = self.instances.get(remote_ip._instance_backend_id)
                # Check if internal IP is pending.
                if instance and not local_ip:
                    local_ip = self.pending_ips.get((instance.backend_id, subnet.backend_id))

            # Create local internal IP if it does not exist yet.
            if local_ip is None:
                local_ip = remote_ip
                local_ip.subnet = subnet
                local_ip.settings = subnet.settings
                local_ip.instance = instance
                local_ip.save()
            else:
                if local_ip.instance != instance:
                    logger.info('About to reassign internal IP from %s to %s', local_ip.instance, instance)
                    local_ip.instance = instance
                    local_ip.save()

                # Update backend ID for pending internal IP.
                update_pulled_fields(local_ip, remote_ip,
                                     models.InternalIP.get_backend_fields() + ('backend_id',))

        if self.stale_ips:
            logger.info('About to remove stale internal IPs: %s', self.stale_ips)
            models.InternalIP.objects.filter(pk__in=self.stale_ips).delete()


class OpenStackTenantBackend(BaseOpenStackBackend):

    DEFAULTS = {
        'console_type': 'novnc',
    }

    def __init__(self, settings):
        super(OpenStackTenantBackend, self).__init__(settings, settings.options['tenant_id'])

    @property
    def external_network_id(self):
        return self.settings.options['external_network_id']

    def pull_service_properties(self):
        self.pull_flavors()
        self.pull_images()
        self.pull_security_groups()
        self.pull_quotas()
        self.pull_networks()
        self.pull_subnets()
        self.pull_internal_ips()
        self.pull_floating_ips()
        self.pull_volume_types()
        self.pull_volume_availability_zones()
        self.pull_instance_availability_zones()

    def pull_resources(self):
        self.pull_volumes()
        self.pull_snapshots()
        self.pull_instances()

    def pull_volumes(self):
        backend_volumes = self.get_volumes()
        volumes = models.Volume.objects.filter(
            service_project_link__service__settings=self.settings,
            state__in=[models.Volume.States.OK, models.Volume.States.ERRED]
        )
        backend_volumes_map = {backend_volume.backend_id: backend_volume for backend_volume in backend_volumes}
        for volume in volumes:
            try:
                backend_volume = backend_volumes_map[volume.backend_id]
            except KeyError:
                handle_resource_not_found(volume)
            else:
                update_pulled_fields(volume, backend_volume, models.Volume.get_backend_fields())
                handle_resource_update_success(volume)

    def pull_snapshots(self):
        backend_snapshots = self.get_snapshots()
        snapshots = models.Snapshot.objects.filter(
            service_project_link__service__settings=self.settings,
            state__in=[models.Snapshot.States.OK, models.Snapshot.States.ERRED])
        backend_snapshots_map = {backend_snapshot.backend_id: backend_snapshot
                                 for backend_snapshot in backend_snapshots}
        for snapshot in snapshots:
            try:
                backend_snapshot = backend_snapshots_map[snapshot.backend_id]
            except KeyError:
                handle_resource_not_found(snapshot)
            else:
                update_pulled_fields(snapshot, backend_snapshot, models.Snapshot.get_backend_fields())
                handle_resource_update_success(snapshot)

    def pull_instances(self):
        backend_instances = self.get_instances()
        instances = models.Instance.objects.filter(
            service_project_link__service__settings=self.settings,
            state__in=[models.Instance.States.OK, models.Instance.States.ERRED],
        )
        backend_instances_map = {backend_instance.backend_id: backend_instance
                                 for backend_instance in backend_instances}
        for instance in instances:
            try:
                backend_instance = backend_instances_map[instance.backend_id]
            except KeyError:
                handle_resource_not_found(instance)
            else:
                self.update_instance_fields(instance, backend_instance)
                # XXX: can be optimized after https://goo.gl/BZKo8Y will be resolved.
                self.pull_instance_security_groups(instance)
                handle_resource_update_success(instance)

    def update_instance_fields(self, instance, backend_instance):
        # Preserve flavor fields in Waldur database if flavor is deleted in OpenStack
        fields = set(models.Instance.get_backend_fields())
        flavor_fields = {'flavor_name', 'flavor_disk', 'ram', 'cores', 'disk'}
        if not backend_instance.flavor_name:
            fields = fields - flavor_fields
        fields = list(fields)

        update_pulled_fields(instance, backend_instance, fields)

    def pull_flavors(self):
        nova = self.nova_client
        try:
            flavors = nova.flavors.findall()
        except nova_exceptions.ClientException as e:
            reraise(e)

        flavor_exclude_regex = self.settings.options.get('flavor_exclude_regex', '')
        name_pattern = re.compile(flavor_exclude_regex) if flavor_exclude_regex else None
        with transaction.atomic():
            cur_flavors = self._get_current_properties(models.Flavor)
            for backend_flavor in flavors:
                if name_pattern is not None and name_pattern.match(backend_flavor.name) is not None:
                    logger.debug('Skipping pull of %s flavor as it matches %s regex pattern.',
                                 backend_flavor.name, flavor_exclude_regex)
                    continue

                cur_flavors.pop(backend_flavor.id, None)
                models.Flavor.objects.update_or_create(
                    settings=self.settings,
                    backend_id=backend_flavor.id,
                    defaults={
                        'name': backend_flavor.name,
                        'cores': backend_flavor.vcpus,
                        'ram': backend_flavor.ram,
                        'disk': self.gb2mb(backend_flavor.disk),
                    })

            models.Flavor.objects.filter(backend_id__in=cur_flavors.keys(), settings=self.settings).delete()

    def pull_images(self):
        self._pull_images(models.Image)

    def pull_floating_ips(self):
        # method assumes that instance internal IPs is up to date.
        neutron = self.neutron_client

        try:
            backend_floating_ips = neutron.list_floatingips(tenant_id=self.tenant_id)['floatingips']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        # Step 1. Prepare data
        imported_ips = {ip.backend_id: ip
                        for ip in (self._backend_floating_ip_to_floating_ip(ip)
                                   for ip in backend_floating_ips)}

        # Floating IP is identified either by backend ID or address
        # Backend ID is preferred way of identification, because ID is unique across the whole deployment.
        # However, initially we're creating new floating IP without backend ID.
        # It is expected that it's backend ID is filled up from actual backend data later.
        # Therefore we need to use address as a transient way of floating IP identification.
        floating_ips = {ip.backend_id: ip for ip in models.FloatingIP.objects.filter(
            settings=self.settings).exclude(backend_id='')}
        floating_ips_map = {ip.address: ip for ip in models.FloatingIP.objects.filter(
            settings=self.settings) if ip.address}
        booked_ips = [backend_id for backend_id, ip in floating_ips.items() if ip.is_booked]

        # all imported IPs except those which are booked.
        ips_to_update = set(imported_ips) - set(booked_ips)

        internal_ips = {ip.backend_id: ip for ip in models.InternalIP.objects.filter(
            subnet__settings=self.settings).exclude(backend_id=None)}

        # Step 2. Update or create imported IPs
        for backend_id in ips_to_update:
            imported_ip = imported_ips[backend_id]
            floating_ip = floating_ips.get(backend_id) or floating_ips_map.get(imported_ip.address)

            internal_ip = internal_ips.get(imported_ip._internal_ip_backend_id)
            if imported_ip._internal_ip_backend_id and internal_ip is None:
                logger.warning('Failed to set internal_ip for Floating IP %s', imported_ip.backend_id)
                continue

            imported_ip.internal_ip = internal_ip
            if not floating_ip:
                imported_ip.save()
            else:
                fields_to_update = models.FloatingIP.get_backend_fields() + ('internal_ip',)
                if floating_ip.address != floating_ip.name:
                    # Don't update user defined name.
                    fields_to_update = [field for field in fields_to_update if field != 'name']

                update_pulled_fields(floating_ip, imported_ip, fields_to_update)

        # Step 3. Delete stale IPs
        ips_to_delete = set(floating_ips) - set(imported_ips)
        if ips_to_delete:
            logger.info('About to delete stale floating IPs: %s', ips_to_delete)
            models.FloatingIP.objects.filter(settings=self.settings,
                                             backend_id__in=ips_to_delete).delete()

    def _backend_floating_ip_to_floating_ip(self, backend_floating_ip, **kwargs):
        floating_ip = models.FloatingIP(
            settings=self.settings,
            name=backend_floating_ip['floating_ip_address'],
            address=backend_floating_ip['floating_ip_address'],
            backend_network_id=backend_floating_ip['floating_network_id'],
            runtime_state=backend_floating_ip['status'],
            backend_id=backend_floating_ip['id'],
        )
        for field, value in kwargs.items():
            setattr(floating_ip, field, value)

        if 'internal_ip' not in kwargs:
            floating_ip._internal_ip_backend_id = backend_floating_ip['port_id']

        return floating_ip

    def _delete_stale_properties(self, model, backend_items):
        with transaction.atomic():
            current_items = model.objects.filter(settings=self.settings)
            current_ids = set(current_items.values_list('backend_id', flat=True))
            backend_ids = set(item['id'] for item in backend_items)
            stale_ids = current_ids - backend_ids
            if stale_ids:
                model.objects.filter(settings=self.settings, backend_id__in=stale_ids).delete()

    def pull_security_groups(self):
        neutron = self.neutron_client
        try:
            security_groups = neutron.list_security_groups(tenant_id=self.tenant_id)['security_groups']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        for backend_security_group in security_groups:
            backend_id = backend_security_group['id']
            defaults = {
                'name': backend_security_group['name'],
                'description': backend_security_group['description']
            }
            try:
                with transaction.atomic():
                    security_group, _ = models.SecurityGroup.objects.update_or_create(
                        settings=self.settings,
                        backend_id=backend_id,
                        defaults=defaults
                    )
                    self._extract_security_group_rules(security_group, backend_security_group)
            except IntegrityError:
                logger.warning('Could not create security group with backend ID %s '
                               'and service settings %s due to concurrent update.',
                               backend_id, self.settings)

        self._delete_stale_properties(models.SecurityGroup, security_groups)

    def pull_quotas(self):
        self._pull_tenant_quotas(self.tenant_id, self.settings)

    def pull_networks(self):
        neutron = self.neutron_client
        try:
            networks = neutron.list_networks(tenant_id=self.tenant_id)['networks']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        for backend_network in networks:
            defaults = {
                'name': backend_network['name'],
                'description': backend_network['description'],
            }
            if backend_network.get('provider:network_type'):
                defaults['type'] = backend_network['provider:network_type']
            if backend_network.get('provider:segmentation_id'):
                defaults['segmentation_id'] = backend_network['provider:segmentation_id']
            backend_id = backend_network['id']
            try:
                models.Network.objects.update_or_create(
                    settings=self.settings,
                    backend_id=backend_id,
                    defaults=defaults
                )
            except IntegrityError:
                logger.warning('Could not create network with backend ID %s '
                               'and service settings %s due to concurrent update.',
                               backend_id, self.settings)

        self._delete_stale_properties(models.Network, networks)

    def pull_subnets(self):
        neutron = self.neutron_client
        try:
            subnets = neutron.list_subnets(tenant_id=self.tenant_id)['subnets']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        current_networks = {
            network.backend_id: network.id
            for network in models.Network.objects.filter(settings=self.settings).only('id', 'backend_id')
        }
        subnet_networks = set(subnet['network_id'] for subnet in subnets)
        missing_networks = subnet_networks - set(current_networks.keys())
        if missing_networks:
            logger.warning('Cannot pull subnets for networks with id %s '
                           'because their networks are not pulled yet.', missing_networks)

        for backend_subnet in subnets:
            backend_id = backend_subnet['id']
            network_id = current_networks.get(backend_subnet['network_id'])
            if not network_id:
                continue
            defaults = {
                'name': backend_subnet['name'],
                'description': backend_subnet['description'],
                'allocation_pools': backend_subnet['allocation_pools'],
                'cidr': backend_subnet['cidr'],
                'ip_version': backend_subnet['ip_version'],
                'gateway_ip': backend_subnet.get('gateway_ip'),
                'enable_dhcp': backend_subnet.get('enable_dhcp', False),
                'network_id': network_id,
            }
            try:
                models.SubNet.objects.update_or_create(
                    settings=self.settings,
                    backend_id=backend_id,
                    defaults=defaults
                )
            except IntegrityError:
                logger.warning('Could not create subnet with backend ID %s '
                               'and service settings %s due to concurrent update.',
                               backend_id, self.settings)

        self._delete_stale_properties(models.SubNet, subnets)

    @log_backend_action()
    def create_volume(self, volume):
        kwargs = {
            'size': self.mb2gb(volume.size),
            'name': volume.name,
            'description': volume.description,
        }

        if volume.source_snapshot:
            kwargs['snapshot_id'] = volume.source_snapshot.backend_id

        tenant = volume.service_project_link.service.settings.scope

        if volume.type:
            kwargs['volume_type'] = volume.type.backend_id
        else:
            volume_type_name = (tenant and tenant.default_volume_type_name)
            if volume_type_name:
                try:
                    volume_type = models.VolumeType.objects.get(name=volume_type_name,
                                                                settings=volume.service_project_link.service.settings)
                    volume.type = volume_type
                    kwargs['volume_type'] = volume_type.backend_id
                except models.VolumeType.DoesNotExist:
                    logger.error('Volume type is not set as volume type with name %s is not found. Settings UUID: %s' %
                                 (volume_type_name, volume.service_project_link.service.settings.uuid.hex))
                except models.VolumeType.MultipleObjectsReturned:
                    logger.error('Volume type is not set as multiple volume types with name %s are found.'
                                 'Service settings UUID: %s' %
                                 (volume_type_name,
                                  volume.service_project_link.service.settings.uuid.hex))

        if volume.availability_zone:
            kwargs['availability_zone'] = volume.availability_zone.name
        else:
            volume_availability_zone_name = \
                (tenant and tenant.service_settings.options.get('volume_availability_zone_name'))

            if volume_availability_zone_name:
                try:
                    volume_availability_zone = models.VolumeAvailabilityZone.objects.get(
                        name=volume_availability_zone_name,
                        settings=volume.service_project_link.service.settings)
                    volume.availability_zone = volume_availability_zone
                    kwargs['availability_zone'] = volume_availability_zone.name
                except models.VolumeAvailabilityZone.DoesNotExist:
                    logger.error('Volume availability zone with name %s is not found. Settings UUID: %s' %
                                 (volume_availability_zone_name, volume.service_project_link.service.settings.uuid.hex))

        if volume.image:
            kwargs['imageRef'] = volume.image.backend_id
        cinder = self.cinder_client
        try:
            backend_volume = cinder.volumes.create(**kwargs)
        except cinder_exceptions.ClientException as e:
            reraise(e)
        volume.backend_id = backend_volume.id
        if hasattr(backend_volume, 'volume_image_metadata'):
            volume.image_metadata = backend_volume.volume_image_metadata
        volume.bootable = backend_volume.bootable == 'true'
        volume.runtime_state = backend_volume.status
        volume.save()
        return volume

    @log_backend_action()
    def update_volume(self, volume):
        cinder = self.cinder_client
        try:
            cinder.volumes.update(volume.backend_id, name=volume.name, description=volume.description)
        except cinder_exceptions.ClientException as e:
            reraise(e)

    @log_backend_action()
    def delete_volume(self, volume):
        cinder = self.cinder_client
        try:
            cinder.volumes.delete(volume.backend_id)
        except cinder_exceptions.NotFound:
            logger.info('OpenStack volume %s has been already deleted', volume.backend_id)
        except cinder_exceptions.ClientException as e:
            reraise(e)
        volume.decrease_backend_quotas_usage()

    @log_backend_action()
    def attach_volume(self, volume, instance_uuid, device=None):
        instance = models.Instance.objects.get(uuid=instance_uuid)
        nova = self.nova_client
        try:
            nova.volumes.create_server_volume(instance.backend_id, volume.backend_id,
                                              device=None if device == '' else device)
        except nova_exceptions.ClientException as e:
            reraise(e)
        else:
            volume.instance = instance
            volume.device = device
            volume.save(update_fields=['instance', 'device'])

    @log_backend_action()
    def detach_volume(self, volume):
        nova = self.nova_client
        try:
            nova.volumes.delete_server_volume(volume.instance.backend_id, volume.backend_id)
        except nova_exceptions.ClientException as e:
            reraise(e)
        else:
            volume.instance = None
            volume.device = ''
            volume.save(update_fields=['instance', 'device'])

    @log_backend_action()
    def extend_volume(self, volume):
        cinder = self.cinder_client
        try:
            cinder.volumes.extend(volume.backend_id, self.mb2gb(volume.size))
        except cinder_exceptions.ClientException as e:
            reraise(e)

    def import_volume(self, backend_volume_id, save=True, service_project_link=None):
        """ Restore Waldur volume instance based on backend data. """
        cinder = self.cinder_client
        try:
            backend_volume = cinder.volumes.get(backend_volume_id)
        except cinder_exceptions.ClientException as e:
            reraise(e)
        volume = self._backend_volume_to_volume(backend_volume)
        if service_project_link is not None:
            volume.service_project_link = service_project_link
        if save:
            volume.save()

        return volume

    def _backend_volume_to_volume(self, backend_volume):
        volume_type = None
        availability_zone = None

        try:
            if backend_volume.volume_type:
                volume_type = models.VolumeType.objects.get(name=backend_volume.volume_type, settings=self.settings)
        except models.VolumeType.DoesNotExist:
            pass
        except models.VolumeType.MultipleObjectsReturned:
            logger.error('Volume type is not set as multiple volume types with name %s are found.'
                         'Service settings UUID: %s',
                         (backend_volume.volume_type,
                          self.settings.uuid.hex))

        try:
            backend_volume_availability_zone = getattr(backend_volume, 'availability_zone', None)
            if backend_volume_availability_zone:
                availability_zone = models.VolumeAvailabilityZone.objects.get(
                    name=backend_volume_availability_zone,
                    settings=self.settings)
        except models.VolumeAvailabilityZone.DoesNotExist:
            pass

        volume = models.Volume(
            name=backend_volume.name,
            description=backend_volume.description or '',
            size=self.gb2mb(backend_volume.size),
            metadata=backend_volume.metadata,
            backend_id=backend_volume.id,
            type=volume_type,
            bootable=backend_volume.bootable == 'true',
            runtime_state=backend_volume.status,
            state=models.Volume.States.OK,
            availability_zone=availability_zone,
        )
        if getattr(backend_volume, 'volume_image_metadata', False):
            volume.image_metadata = backend_volume.volume_image_metadata
            try:
                image_id = volume.image_metadata.get('image_id')
                if image_id:
                    volume.image = models.Image.objects.get(
                        settings=self.settings, backend_id=image_id)
            except models.Image.DoesNotExist:
                pass
        # In our setup volume could be attached only to one instance.
        if getattr(backend_volume, 'attachments', False):
            if 'device' in backend_volume.attachments[0]:
                volume.device = backend_volume.attachments[0]['device']

            if 'server_id' in backend_volume.attachments[0]:
                volume.instance = models.Instance.objects.filter(
                    service_project_link__service__settings=self.settings,
                    backend_id=backend_volume.attachments[0]['server_id'],
                ).first()
        return volume

    def get_volumes(self):
        cinder = self.cinder_client
        try:
            backend_volumes = cinder.volumes.list()
        except cinder_exceptions.ClientException as e:
            reraise(e)
        return [self._backend_volume_to_volume(backend_volume) for backend_volume in backend_volumes]

    def get_volumes_for_import(self):
        return self._get_backend_resource(models.Volume, self.get_volumes())

    @log_backend_action()
    def remove_bootable_flag(self, volume):
        cinder = self.cinder_client
        try:
            backend_volume = cinder.volumes.get(volume.backend_id)
            cinder.volumes.set_bootable(backend_volume, False)
        except cinder_exceptions.ClientException as e:
            reraise(e)
        volume.bootable = False
        volume.save(update_fields=['bootable'])

    @log_backend_action()
    def pull_volume(self, volume, update_fields=None):
        import_time = timezone.now()
        imported_volume = self.import_volume(volume.backend_id, save=False)

        volume.refresh_from_db()
        if volume.modified < import_time:
            if not update_fields:
                update_fields = models.Volume.get_backend_fields()

            update_pulled_fields(volume, imported_volume, update_fields)

    @log_backend_action()
    def pull_volume_runtime_state(self, volume):
        cinder = self.cinder_client
        try:
            backend_volume = cinder.volumes.get(volume.backend_id)
        except cinder_exceptions.NotFound:
            volume.runtime_state = 'deleted'
            volume.save(update_fields=['runtime_state'])
            return
        except cinder_exceptions.ClientException as e:
            reraise(e)
        else:
            if backend_volume.status != volume.runtime_state:
                volume.runtime_state = backend_volume.status
                volume.save(update_fields=['runtime_state'])

    @log_backend_action('check is volume deleted')
    def is_volume_deleted(self, volume):
        cinder = self.cinder_client
        try:
            cinder.volumes.get(volume.backend_id)
            return False
        except cinder_exceptions.NotFound:
            return True
        except cinder_exceptions.ClientException as e:
            reraise(e)

    @log_backend_action()
    def create_snapshot(self, snapshot, force=True):
        kwargs = {
            'name': snapshot.name,
            'description': snapshot.description,
            'force': force,
        }
        cinder = self.cinder_client
        try:
            backend_snapshot = cinder.volume_snapshots.create(snapshot.source_volume.backend_id, **kwargs)
        except cinder_exceptions.ClientException as e:
            reraise(e)
        snapshot.backend_id = backend_snapshot.id
        snapshot.runtime_state = backend_snapshot.status
        snapshot.size = self.gb2mb(backend_snapshot.size)
        snapshot.save()
        return snapshot

    def import_snapshot(self, backend_snapshot_id, save=True, service_project_link=None):
        """ Restore Waldur Snapshot instance based on backend data. """
        cinder = self.cinder_client
        try:
            backend_snapshot = cinder.volume_snapshots.get(backend_snapshot_id)
        except cinder_exceptions.ClientException as e:
            reraise(e)
        snapshot = self._backend_snapshot_to_snapshot(backend_snapshot)
        if service_project_link is not None:
            snapshot.service_project_link = service_project_link
        if save:
            snapshot.save()
        return snapshot

    def _backend_snapshot_to_snapshot(self, backend_snapshot):
        snapshot = models.Snapshot(
            name=backend_snapshot.name,
            description=backend_snapshot.description or '',
            size=self.gb2mb(backend_snapshot.size),
            metadata=backend_snapshot.metadata,
            backend_id=backend_snapshot.id,
            runtime_state=backend_snapshot.status,
            state=models.Snapshot.States.OK,
        )
        if hasattr(backend_snapshot, 'volume_id'):
            snapshot.source_volume = models.Volume.objects.filter(
                service_project_link__service__settings=self.settings, backend_id=backend_snapshot.volume_id).first()
        return snapshot

    def get_snapshots(self):
        cinder = self.cinder_client
        try:
            backend_snapshots = cinder.volume_snapshots.list()
        except cinder_exceptions.ClientException as e:
            reraise(e)
        return [self._backend_snapshot_to_snapshot(backend_snapshot) for backend_snapshot in backend_snapshots]

    def get_snapshots_for_import(self):
        return self._get_backend_resource(models.Snapshot, self.get_snapshots())

    @log_backend_action()
    def pull_snapshot(self, snapshot, update_fields=None):
        import_time = timezone.now()
        imported_snapshot = self.import_snapshot(snapshot.backend_id, save=False)

        snapshot.refresh_from_db()
        if snapshot.modified < import_time:
            if update_fields is None:
                update_fields = models.Snapshot.get_backend_fields()
            update_pulled_fields(snapshot, imported_snapshot, update_fields)

    @log_backend_action()
    def pull_snapshot_runtime_state(self, snapshot):
        cinder = self.cinder_client
        try:
            backend_snapshot = cinder.volume_snapshots.get(snapshot.backend_id)
        except cinder_exceptions.ClientException as e:
            reraise(e)
        if backend_snapshot.status != snapshot.runtime_state:
            snapshot.runtime_state = backend_snapshot.status
            snapshot.save(update_fields=['runtime_state'])
        return snapshot

    @log_backend_action()
    def delete_snapshot(self, snapshot):
        cinder = self.cinder_client
        try:
            cinder.volume_snapshots.delete(snapshot.backend_id)
        except cinder_exceptions.ClientException as e:
            reraise(e)
        snapshot.decrease_backend_quotas_usage()

    @log_backend_action()
    def update_snapshot(self, snapshot):
        cinder = self.cinder_client
        try:
            cinder.volume_snapshots.update(
                snapshot.backend_id, name=snapshot.name, description=snapshot.description)
        except cinder_exceptions.ClientException as e:
            reraise(e)

    @log_backend_action('check is snapshot deleted')
    def is_snapshot_deleted(self, snapshot):
        cinder = self.cinder_client
        try:
            cinder.volume_snapshots.get(snapshot.backend_id)
            return False
        except cinder_exceptions.NotFound:
            return True
        except cinder_exceptions.ClientException as e:
            reraise(e)

    def is_volume_availability_zone_supported(self):
        return 'AvailabilityZones' in [e.name for e in list_extensions.ListExtManager(self.cinder_client).show_all()]

    @log_backend_action()
    def create_instance(self, instance, backend_flavor_id=None, public_key=None):
        nova = self.nova_client

        try:
            backend_flavor = nova.flavors.get(backend_flavor_id)

            # instance key name and fingerprint are optional
            # it is assumed that if public_key is specified, then
            # key_name and key_fingerprint have valid values
            if public_key:
                backend_public_key = self._get_or_create_ssh_key(
                    instance.key_name, instance.key_fingerprint, public_key)
            else:
                backend_public_key = None

            try:
                instance.volumes.get(bootable=True)
            except models.Volume.DoesNotExist:
                raise OpenStackBackendError(
                    'Current installation cannot create instance without a system volume.')

            nics = [{'port-id': internal_ip.backend_id}
                    for internal_ip in instance.internal_ips_set.all()
                    if internal_ip.backend_id]

            block_device_mapping_v2 = []
            for volume in instance.volumes.iterator():
                device_mapping = {
                    'destination_type': 'volume',
                    'device_type': 'disk',
                    'source_type': 'volume',
                    'uuid': volume.backend_id,
                    'delete_on_termination': True,
                }
                if volume.bootable:
                    device_mapping.update({'boot_index': 0})

                block_device_mapping_v2.append(device_mapping)

            server_create_parameters = dict(
                name=instance.name,
                image=None,  # Boot from volume, see boot_index above
                flavor=backend_flavor,
                block_device_mapping_v2=block_device_mapping_v2,
                nics=nics,
                key_name=backend_public_key.name if backend_public_key is not None else None,
            )
            if instance.availability_zone:
                server_create_parameters['availability_zone'] = instance.availability_zone.name
            else:
                availability_zone = self.settings.options.get('availability_zone')
                if availability_zone:
                    server_create_parameters['availability_zone'] = availability_zone

            if instance.user_data:
                server_create_parameters['userdata'] = instance.user_data

            server = nova.servers.create(**server_create_parameters)
            instance.backend_id = server.id
            instance.save()
        except nova_exceptions.ClientException as e:
            logger.exception("Failed to provision instance %s", instance.uuid)
            reraise(e)
        else:
            logger.info("Successfully provisioned instance %s", instance.uuid)

    @log_backend_action()
    def pull_instance_floating_ips(self, instance):
        # method assumes that instance internal IPs is up to date.
        neutron = self.neutron_client

        internal_ip_mappings = {ip.backend_id: ip for ip in instance.internal_ips_set.all().exclude(backend_id=None)}
        try:
            backend_floating_ips = neutron.list_floatingips(
                tenant_id=self.tenant_id, port_id=internal_ip_mappings.keys())['floatingips']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        backend_ids = {fip['id'] for fip in backend_floating_ips}

        floating_ips = {
            fip.backend_id: fip
            for fip in models.FloatingIP.objects.filter(settings=self.settings,
                                                        is_booked=False,
                                                        backend_id__in=backend_ids)
        }

        with transaction.atomic():
            for backend_floating_ip in backend_floating_ips:
                imported_floating_ip = self._backend_floating_ip_to_floating_ip(backend_floating_ip)
                internal_ip = internal_ip_mappings.get(imported_floating_ip._internal_ip_backend_id)

                floating_ip = floating_ips.get(imported_floating_ip.backend_id)
                if floating_ip is None:
                    imported_floating_ip.internal_ip = internal_ip
                    imported_floating_ip.save()
                    continue

                if floating_ip.internal_ip != internal_ip:
                    floating_ip.internal_ip = internal_ip
                    floating_ip.save()

                # Don't update user defined name.
                if floating_ip.address != floating_ip.name:
                    imported_floating_ip.name = floating_ip.name
                update_pulled_fields(floating_ip, imported_floating_ip, models.FloatingIP.get_backend_fields())

            frontend_ids = set(instance.floating_ips.filter(is_booked=False)
                               .exclude(backend_id='')
                               .values_list('backend_id', flat=True))
            stale_ids = frontend_ids - backend_ids
            logger.info('About to detach floating IPs from internal IPs: %s', stale_ids)
            instance.floating_ips.filter(backend_id__in=stale_ids).update(internal_ip=None)

    @log_backend_action()
    def push_instance_floating_ips(self, instance):
        neutron = self.neutron_client
        instance_floating_ips = instance.floating_ips
        try:
            backend_floating_ips = neutron.list_floatingips(
                port_id=instance.internal_ips_set.values_list('backend_id', flat=True))['floatingips']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        # disconnect stale
        instance_floating_ips_ids = [fip.backend_id for fip in instance_floating_ips]
        for backend_floating_ip in backend_floating_ips:
            if backend_floating_ip['id'] not in instance_floating_ips_ids:
                try:
                    neutron.update_floatingip(backend_floating_ip['id'], body={'floatingip': {'port_id': None}})
                except neutron_exceptions.NeutronClientException as e:
                    reraise(e)

        # connect new ones
        backend_floating_ip_ids = {fip['id']: fip for fip in backend_floating_ips}
        for floating_ip in instance_floating_ips:
            backend_floating_ip = backend_floating_ip_ids.get(floating_ip.backend_id)
            if not backend_floating_ip or backend_floating_ip['port_id'] != floating_ip.internal_ip.backend_id:
                try:
                    neutron.update_floatingip(
                        floating_ip.backend_id, body={'floatingip': {'port_id': floating_ip.internal_ip.backend_id}})
                except neutron_exceptions.NeutronClientException as e:
                    reraise(e)

    def create_floating_ip(self, floating_ip):
        neutron = self.neutron_client
        try:
            backend_floating_ip = neutron.create_floatingip({
                'floatingip': {
                    'floating_network_id': floating_ip.backend_network_id,
                    'tenant_id': self.tenant_id,
                }
            })['floatingip']
        except neutron_exceptions.NeutronClientException as e:
            floating_ip.runtime_state = 'ERRED'
            floating_ip.save()
            reraise(e)
        else:
            floating_ip.address = backend_floating_ip['floating_ip_address']
            floating_ip.backend_id = backend_floating_ip['id']
            floating_ip.save()

    @log_backend_action()
    def delete_floating_ip(self, floating_ip):
        self._delete_backend_floating_ip(floating_ip.backend_id, self.tenant_id)
        floating_ip.decrease_backend_quotas_usage()
        floating_ip.delete()

    @log_backend_action()
    def pull_floating_ip_runtime_state(self, floating_ip):
        neutron = self.neutron_client
        try:
            backend_floating_ip = neutron.show_floatingip(floating_ip.backend_id)['floatingip']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)
        else:
            floating_ip.runtime_state = backend_floating_ip['status']
            floating_ip.save()

    def _get_or_create_ssh_key(self, key_name, fingerprint, public_key):
        nova = self.nova_client

        try:
            return nova.keypairs.find(fingerprint=fingerprint)
        except nova_exceptions.NotFound:
            # Fine, it's a new key, let's add it
            try:
                logger.info('Propagating ssh public key %s to backend', key_name)
                return nova.keypairs.create(name=key_name, public_key=public_key)
            except nova_exceptions.ClientException as e:
                logger.error('Unable to import SSH public key to OpenStack, '
                             'key_name: %s, fingerprint: %s, public_key: %s, error: %s',
                             key_name, fingerprint, public_key, e)
                reraise(e)
        except nova_exceptions.ClientException as e:
            reraise(e)

    @log_backend_action()
    def update_instance(self, instance):
        nova = self.nova_client
        try:
            nova.servers.update(instance.backend_id, name=instance.name)
        except keystone_exceptions.NotFound as e:
            reraise(e)

    def import_instance(self, backend_instance_id, save=True, service_project_link=None):
        # NB! This method does not import instance sub-objects like security groups or internal IPs.
        #     They have to be pulled separately.
        nova = self.nova_client
        try:
            backend_instance = nova.servers.get(backend_instance_id)
            attached_volume_ids = [v.volumeId for v in nova.volumes.get_server_volumes(backend_instance_id)]
            flavor_id = backend_instance.flavor['id']
        except nova_exceptions.ClientException as e:
            reraise(e)

        instance = self._backend_instance_to_instance(backend_instance, flavor_id)
        with transaction.atomic():
            # import instance volumes, or use existed if they already exist in Waldur.
            volumes = []
            for backend_volume_id in attached_volume_ids:
                try:
                    volumes.append(models.Volume.objects.get(
                        service_project_link__service__settings=self.settings, backend_id=backend_volume_id))
                except models.Volume.DoesNotExist:
                    volumes.append(self.import_volume(backend_volume_id, save, service_project_link))

            if service_project_link:
                instance.service_project_link = service_project_link
            if hasattr(backend_instance, 'fault'):
                instance.error_message = backend_instance.fault['message']
            if save:
                instance.save()
                instance.volumes.add(*volumes)

        return instance

    def _backend_instance_to_instance(self, backend_instance, backend_flavor_id=None):
        # parse launch time
        try:
            d = dateparse.parse_datetime(backend_instance.to_dict()['OS-SRV-USG:launched_at'])
        except (KeyError, ValueError, TypeError):
            launch_time = None
        else:
            # At the moment OpenStack does not provide any timezone info,
            # but in future it might do.
            if timezone.is_naive(d):
                launch_time = timezone.make_aware(d, timezone.utc)

        availability_zone = None
        try:
            availability_zone_name = backend_instance.to_dict().get('OS-EXT-AZ:availability_zone')
            if availability_zone_name:
                availability_zone = models.InstanceAvailabilityZone.objects.get(
                    name=availability_zone_name, settings=self.settings)
        except (KeyError, ValueError, TypeError, models.InstanceAvailabilityZone.DoesNotExist):
            pass

        instance = models.Instance(
            name=backend_instance.name or backend_instance.id,
            key_name=backend_instance.key_name or '',
            start_time=launch_time,
            state=models.Instance.States.OK,
            runtime_state=backend_instance.status,
            created=dateparse.parse_datetime(backend_instance.created),
            backend_id=backend_instance.id,
            availability_zone=availability_zone,
        )

        if backend_flavor_id:
            try:
                flavor = models.Flavor.objects.get(settings=self.settings, backend_id=backend_flavor_id)
                instance.flavor_name = flavor.name
                instance.flavor_disk = flavor.disk
                instance.cores = flavor.cores
                instance.ram = flavor.ram
            except models.Flavor.DoesNotExist:
                backend_flavor = self._get_flavor(backend_flavor_id)
                # If flavor has been removed in OpenStack cloud, we should skip update
                if backend_flavor:
                    instance.flavor_name = backend_flavor.name
                    instance.flavor_disk = self.gb2mb(backend_flavor.disk)
                    instance.cores = backend_flavor.vcpus
                    instance.ram = backend_flavor.ram

        return instance

    def _get_flavor(self, flavor_id):
        nova = self.nova_client
        try:
            return nova.flavors.get(flavor_id)
        except nova_exceptions.NotFound:
            logger.info('OpenStack flavor %s is gone.', flavor_id)
            return None
        except nova_exceptions.ClientException as e:
            reraise(e)

    def _get_backend_resource(self, model, resources):
        registered_backend_ids = model.objects.filter(
            service_project_link__service__settings=self.settings).values_list('backend_id', flat=True)
        return [instance for instance in resources if instance.backend_id not in registered_backend_ids]

    def get_instances(self):
        nova = self.nova_client
        try:
            backend_instances = nova.servers.list()
        except nova_exceptions.ClientException as e:
            reraise(e)

        instances = []
        for backend_instance in backend_instances:
            flavor_id = backend_instance.flavor['id']
            instances.append(self._backend_instance_to_instance(backend_instance, flavor_id))
        return instances

    def get_instances_for_import(self):
        return self._get_backend_resource(models.Instance, self.get_instances())

    @transaction.atomic()
    def _pull_zones(self, backend_zones, frontend_model, default_zone='nova'):
        """
        This method is called for Volume and Instance Availability zone synchronization.
        It is assumed that default zone could not be used for Volume or Instance provisioning.
        Therefore we do not pull default zone at all. Please note, however, that default zone
        name could be changed in Nova and Cinder config. We don't support this use case either.

        All availability zones are split into 3 subsets: stale, missing and common.
        Stale zone are removed, missing zones are created.
        If zone state has been changed, it is synchronized.
        """
        front_zones_map = {
            zone.name: zone
            for zone in frontend_model.objects.filter(settings=self.settings)
        }

        back_zones_map = {
            zone.zoneName: zone.zoneState.get('available', True)
            for zone in backend_zones
            if zone.zoneName != default_zone
        }

        missing_zones = set(back_zones_map.keys()) - set(front_zones_map.keys())
        for zone in missing_zones:
            frontend_model.objects.create(
                settings=self.settings,
                name=zone,
            )

        stale_zones = set(front_zones_map.keys()) - set(back_zones_map.keys())
        frontend_model.objects.filter(
            name__in=stale_zones,
            settings=self.settings
        ).delete()

        common_zones = set(front_zones_map.keys()) & set(back_zones_map.keys())
        for zone_name in common_zones:
            zone = front_zones_map[zone_name]
            actual = back_zones_map[zone_name]
            if zone.available != actual:
                zone.available = actual
                zone.save(update_fields=['available'])

    def pull_instance_availability_zones(self):
        try:
            # By default detailed flag is True, but OpenStack policy for detailed data is disabled.
            # Therefore we should explicitly pass detailed=False. Otherwise request fails.
            backend_zones = self.nova_client.availability_zones.list(detailed=False)
        except nova_exceptions.ClientException as e:
            reraise(e)

        self._pull_zones(backend_zones, models.InstanceAvailabilityZone)

    @log_backend_action()
    def pull_instance(self, instance, update_fields=None):
        import_time = timezone.now()
        imported_instance = self.import_instance(instance.backend_id, save=False)

        instance.refresh_from_db()
        if instance.modified < import_time:
            if update_fields is None:
                update_fields = models.Instance.get_backend_fields()
            update_pulled_fields(instance, imported_instance, update_fields)

    @log_backend_action()
    def pull_instance_internal_ips(self, instance):
        neutron = self.neutron_client
        try:
            backend_internal_ips = neutron.list_ports(device_id=instance.backend_id)['ports']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        existing_ips = {
            ip.backend_id: ip
            for ip in instance.internal_ips_set.exclude(backend_id=None)
        }

        pending_ips = {
            ip.subnet.backend_id: ip
            for ip in instance.internal_ips_set.filter(backend_id=None)
        }

        local_ips = {
            ip.backend_id: ip
            for ip in (models.InternalIP.objects.filter(subnet__settings=self.settings)
                       .exclude(backend_id=None))
        }

        subnets = models.SubNet.objects.filter(settings=self.settings)
        subnet_mappings = {subnet.backend_id: subnet for subnet in subnets}

        with transaction.atomic():
            for backend_internal_ip in backend_internal_ips:
                imported_internal_ip = backend_internal_ip_to_internal_ip(backend_internal_ip, instance=instance)
                subnet = subnet_mappings.get(imported_internal_ip._subnet_backend_id)
                if subnet is None:
                    logger.warning('Skipping Neutron port synchronization process because '
                                   'related subnet is not imported yet. Port ID: %s, subnet ID: %s',
                                   imported_internal_ip.backend_id, imported_internal_ip._subnet_backend_id)
                    continue

                if imported_internal_ip._subnet_backend_id in pending_ips:
                    internal_ip = pending_ips[imported_internal_ip._subnet_backend_id]
                    # Update backend ID for pending internal IP
                    update_pulled_fields(internal_ip, imported_internal_ip,
                                         models.InternalIP.get_backend_fields() + ('backend_id',))

                elif imported_internal_ip.backend_id in existing_ips:
                    internal_ip = existing_ips[imported_internal_ip.backend_id]
                    update_pulled_fields(internal_ip, imported_internal_ip, models.InternalIP.get_backend_fields())

                elif imported_internal_ip.backend_id in local_ips:
                    internal_ip = local_ips[imported_internal_ip.backend_id]
                    if internal_ip.instance != instance:
                        logger.info('About to reassign shared internal IP from instance %s to instance %s',
                                    internal_ip.instance, instance)
                        internal_ip.instance = instance
                        internal_ip.save()

                else:
                    logger.debug('About to create internal IP. Instance ID: %s, subnet ID: %s',
                                 instance.backend_id, subnet.backend_id)
                    internal_ip = imported_internal_ip
                    internal_ip.subnet = subnet
                    internal_ip.settings = subnet.settings
                    internal_ip.instance = instance
                    internal_ip.save()

            # remove stale internal IPs
            frontend_ids = set(existing_ips.keys())
            backend_ids = {internal_ip['id'] for internal_ip in backend_internal_ips}
            stale_ids = frontend_ids - backend_ids
            if stale_ids:
                logger.info('About to delete internal IPs with IDs %s', stale_ids)
                instance.internal_ips_set.filter(backend_id__in=stale_ids).delete()

    def pull_internal_ips(self):
        synchronizer = InternalIPSynchronizer(self.neutron_client, self.tenant_id, self.settings)
        synchronizer.execute()

    @log_backend_action()
    def push_instance_internal_ips(self, instance):
        # we assume that internal IP subnet cannot be changed
        neutron = self.neutron_client
        nova = self.nova_client

        try:
            backend_internal_ips = neutron.list_ports(device_id=instance.backend_id)['ports']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        # delete stale internal IPs
        exist_ids = instance.internal_ips_set.values_list('backend_id', flat=True)
        for backend_internal_ip in backend_internal_ips:
            if backend_internal_ip['id'] not in exist_ids:
                try:
                    logger.info('About to delete network port with ID %s.', backend_internal_ip['id'])
                    neutron.delete_port(backend_internal_ip['id'])
                except neutron_exceptions.NeutronClientException as e:
                    reraise(e)

        # create new internal IPs
        new_internal_ips = instance.internal_ips_set.exclude(backend_id__in=[ip['id'] for ip in backend_internal_ips])
        for new_internal_ip in new_internal_ips:
            try:
                port = {
                    'network_id': new_internal_ip.subnet.network.backend_id,
                    # If you specify only a subnet ID, OpenStack Networking
                    # allocates an available IP from that subnet to the port.
                    'fixed_ips': [{
                        'subnet_id': new_internal_ip.subnet.backend_id,
                    }],
                    'security_groups': list(instance.security_groups.exclude(backend_id='').
                                            values_list('backend_id', flat=True)),
                }
                logger.debug('About to create network port for instance %s in subnet %s.',
                             instance.backend_id, new_internal_ip.subnet.backend_id)
                backend_internal_ip = neutron.create_port({'port': port})['port']
                nova.servers.interface_attach(instance.backend_id, backend_internal_ip['id'], None, None)
            except neutron_exceptions.NeutronClientException as e:
                reraise(e)
            except nova_exceptions.ClientException as e:
                reraise(e)
            new_internal_ip.mac_address = backend_internal_ip['mac_address']
            new_internal_ip.ip4_address = backend_internal_ip['fixed_ips'][0]['ip_address']
            new_internal_ip.backend_id = backend_internal_ip['id']
            new_internal_ip.save()

    @log_backend_action()
    def create_instance_internal_ips(self, instance):
        security_groups = list(instance.security_groups.values_list('backend_id', flat=True))
        for internal_ip in instance.internal_ips_set.all():
            self.create_internal_ip(internal_ip, security_groups)

    def create_internal_ip(self, internal_ip, security_groups):
        neutron = self.neutron_client

        logger.debug('About to create network port. Network ID: %s. Subnet ID: %s.',
                     internal_ip.subnet.network.backend_id, internal_ip.subnet.backend_id)

        try:
            port = {
                'network_id': internal_ip.subnet.network.backend_id,
                'fixed_ips': [{
                    'subnet_id': internal_ip.subnet.backend_id,
                }],
                'security_groups': security_groups,
            }
            backend_internal_ip = neutron.create_port({'port': port})['port']
        except neutron_exceptions.NeutronClientException as e:
            reraise(e)

        internal_ip.mac_address = backend_internal_ip['mac_address']
        internal_ip.ip4_address = backend_internal_ip['fixed_ips'][0]['ip_address']
        internal_ip.backend_id = backend_internal_ip['id']
        internal_ip.save()

    @log_backend_action()
    def delete_instance_internal_ips(self, instance):
        for internal_ip in instance.internal_ips_set.all():
            if internal_ip.backend_id:
                self.delete_internal_ip(internal_ip)

    def delete_internal_ip(self, internal_ip):
        neutron = self.neutron_client

        logger.debug('About to delete network port. Port ID: %s.', internal_ip.backend_id)
        try:
            neutron.delete_port(internal_ip.backend_id)
        except neutron_exceptions.NotFound:
            logger.debug('Neutron port is already deleted. Backend ID: %s.',
                         internal_ip.backend_id)
        except neutron_exceptions.NeutronClientException as e:
            logger.warning('Unable to delete OpenStack network port. '
                           'Skipping error and trying to continue instance deletion. '
                           'Backend ID: %s. Error message is: %s',
                           internal_ip.backend_id, e)
        internal_ip.delete()

    @log_backend_action()
    def pull_instance_security_groups(self, instance):
        nova = self.nova_client
        server_id = instance.backend_id
        try:
            backend_groups = nova.servers.list_security_group(server_id)
        except nova_exceptions.ClientException as e:
            reraise(e)

        backend_ids = set(g.id for g in backend_groups)
        nc_ids = set(
            models.SecurityGroup.objects
            .filter(instances=instance)
            .exclude(backend_id='')
            .values_list('backend_id', flat=True)
        )

        # remove stale groups
        stale_groups = models.SecurityGroup.objects.filter(backend_id__in=(nc_ids - backend_ids))
        instance.security_groups.remove(*stale_groups)

        # add missing groups
        for group_id in backend_ids - nc_ids:
            try:
                security_group = models.SecurityGroup.objects.get(settings=self.settings, backend_id=group_id)
            except models.SecurityGroup.DoesNotExist:
                logger.exception(
                    'Security group with id %s does not exist at Waldur. '
                    'Settings ID: %s' % (group_id, self.settings.id))
            else:
                instance.security_groups.add(security_group)

    @log_backend_action()
    def push_instance_security_groups(self, instance):
        nova = self.nova_client
        server_id = instance.backend_id
        try:
            backend_ids = set(g.id for g in nova.servers.list_security_group(server_id))
        except nova_exceptions.ClientException as e:
            reraise(e)

        nc_ids = set(
            models.SecurityGroup.objects
            .filter(instances=instance)
            .exclude(backend_id='')
            .values_list('backend_id', flat=True)
        )

        # remove stale groups
        for group_id in backend_ids - nc_ids:
            try:
                nova.servers.remove_security_group(server_id, group_id)
            except nova_exceptions.ClientException:
                logger.exception('Failed to remove security group %s from instance %s', group_id, server_id)
            else:
                logger.info('Removed security group %s from instance %s', group_id, server_id)

        # add missing groups
        for group_id in nc_ids - backend_ids:
            try:
                nova.servers.add_security_group(server_id, group_id)
            except nova_exceptions.ClientException:
                logger.exception('Failed to add security group %s to instance %s', group_id, server_id)
            else:
                logger.info('Added security group %s to instance %s', group_id, server_id)

    @log_backend_action()
    def delete_instance(self, instance):
        nova = self.nova_client
        try:
            nova.servers.delete(instance.backend_id)
        except nova_exceptions.NotFound:
            logger.info('OpenStack instance %s is already deleted', instance.backend_id)
        except nova_exceptions.ClientException as e:
            reraise(e)
        instance.decrease_backend_quotas_usage()
        for volume in instance.volumes.all():
            volume.decrease_backend_quotas_usage()

    @log_backend_action('check is instance deleted')
    def is_instance_deleted(self, instance):
        nova = self.nova_client
        try:
            nova.servers.get(instance.backend_id)
            return False
        except nova_exceptions.NotFound:
            return True
        except nova_exceptions.ClientException as e:
            reraise(e)

    @log_backend_action()
    def start_instance(self, instance):
        nova = self.nova_client
        try:
            nova.servers.start(instance.backend_id)
        except nova_exceptions.ClientException as e:
            if e.code == 409 and 'it is in vm_state active' in e.message:
                logger.info('OpenStack instance %s is already started', instance.backend_id)
                return
            reraise(e)

    @log_backend_action()
    def stop_instance(self, instance):
        nova = self.nova_client
        try:
            nova.servers.stop(instance.backend_id)
        except nova_exceptions.ClientException as e:
            if e.code == 409 and 'it is in vm_state stopped' in e.message:
                logger.info('OpenStack instance %s is already stopped', instance.backend_id)
                return
            reraise(e)
        else:
            instance.start_time = None
            instance.save(update_fields=['start_time'])

    @log_backend_action()
    def restart_instance(self, instance):
        nova = self.nova_client
        try:
            nova.servers.reboot(instance.backend_id)
        except nova_exceptions.ClientException as e:
            reraise(e)

    @log_backend_action()
    def resize_instance(self, instance, flavor_id):
        nova = self.nova_client
        try:
            nova.servers.resize(instance.backend_id, flavor_id, 'MANUAL')
        except nova_exceptions.ClientException as e:
            reraise(e)

    @log_backend_action()
    def pull_instance_runtime_state(self, instance):
        nova = self.nova_client
        try:
            backend_instance = nova.servers.get(instance.backend_id)
        except nova_exceptions.ClientException as e:
            reraise(e)
        if backend_instance.status != instance.runtime_state:
            instance.runtime_state = backend_instance.status
            instance.save(update_fields=['runtime_state'])

        if hasattr(backend_instance, 'fault'):
            error_message = backend_instance.fault['message']
            if instance.error_message != error_message:
                instance.error_message = error_message
                instance.save(update_fields=['error_message'])

    @log_backend_action()
    def confirm_instance_resize(self, instance):
        nova = self.nova_client
        try:
            nova.servers.confirm_resize(instance.backend_id)
        except nova_exceptions.ClientException as e:
            reraise(e)

    @log_backend_action()
    def list_meters(self, resource):
        try:
            file_name = self._get_meters_file_name(resource.__class__)
            with open(file_name) as meters_file:
                meters = json.load(meters_file)
        except (KeyError, IOError):
            raise OpenStackBackendError("Cannot find meters for the '%s' resources" % resource.__class__.__name__)

        return meters

    @log_backend_action()
    def get_meter_samples(self, resource, meter_name, start=None, end=None):
        query = [dict(field='resource_id', op='eq', value=resource.backend_id)]

        if start is not None:
            query.append(dict(field='timestamp', op='ge', value=start.strftime('%Y-%m-%dT%H:%M:%S')))
        if end is not None:
            query.append(dict(field='timestamp', op='le', value=end.strftime('%Y-%m-%dT%H:%M:%S')))

        ceilometer = self.ceilometer_client
        try:
            samples = ceilometer.samples.list(meter_name=meter_name, q=query)
        except ceilometer_exceptions.BaseException as e:
            reraise(e)

        return samples

    @log_backend_action()
    def get_console_url(self, instance):
        nova = self.nova_client
        url = None
        console_type = self.settings.get_option('console_type')
        try:
            url = nova.servers.get_console_url(instance.backend_id, console_type)
        except nova_exceptions.ClientException as e:
            reraise(e)

        return url['console']['url']

    @log_backend_action()
    def get_console_output(self, instance, length=None):
        nova = self.nova_client
        try:
            return nova.servers.get_console_output(instance.backend_id, length)
        except nova_exceptions.ClientException as e:
            reraise(e)

    def pull_volume_types(self):
        try:
            volume_types = self.cinder_client.volume_types.list()
        except nova_exceptions.ClientException as e:
            reraise(e)

        with transaction.atomic():
            cur_volume_types = self._get_current_properties(models.VolumeType)
            for backend_type in volume_types:
                cur_volume_types.pop(backend_type.id, None)
                models.VolumeType.objects.update_or_create(
                    settings=self.settings,
                    backend_id=backend_type.id,
                    defaults={
                        'name': backend_type.name,
                        'description': backend_type.description or '',
                    })

            models.VolumeType.objects.filter(backend_id__in=cur_volume_types.keys(), settings=self.settings).delete()

    def pull_volume_availability_zones(self):
        if not self.is_volume_availability_zone_supported():
            return

        try:
            backend_zones = self.cinder_client.availability_zones.list()
        except cinder_exceptions.ClientException as e:
            reraise(e)

        self._pull_zones(backend_zones, models.VolumeAvailabilityZone)



from django.contrib import admin
from . import models


class SlurmPackageAdmin(admin.ModelAdmin):
    list_display = ('name', 'service_settings', 'cpu_price', 'gpu_price', 'ram_price')
    list_filter = ('service_settings',)


admin.site.register(models.SlurmPackage, SlurmPackageAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-03 17:02
from __future__ import unicode_literals

from django.db import migrations
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0023_invoice_current_cost'),
    ]

    operations = [
        migrations.AlterField(
            model_name='genericinvoiceitem',
            name='details',
            field=waldur_core.core.fields.JSONField(blank=True, default=dict, help_text='Stores data about scope'),
        ),
        migrations.AlterField(
            model_name='offeringitem',
            name='offering_details',
            field=waldur_core.core.fields.JSONField(blank=True, default=dict, help_text='Stores data about offering'),
        ),
        migrations.AlterField(
            model_name='openstackitem',
            name='package_details',
            field=waldur_core.core.fields.JSONField(blank=True, default=dict, help_text='Stores data about package'),
        ),
    ]



from waldur_core.core.utils import DryRunCommand
from waldur_mastermind.marketplace_openstack import utils


class Command(DryRunCommand):
    help = """Import OpenStack tenant resources as marketplace resources.
    It is expected that offerings for OpenStack tenant service settings are imported before this command is ran.
    """

    def handle(self, dry_run, *args, **options):
        resources_counter = utils.import_openstack_instances_and_volumes(dry_run)
        self.stdout.write(self.style.SUCCESS('%s resources have been created.' % resources_counter))



from __future__ import unicode_literals

from waldur_core.core import WaldurExtension


class AuthBCCExtension(WaldurExtension):

    class Settings:
        WALDUR_AUTH_BCC = {
            'ENABLED': False,
            'BASE_API_URL': '',
            'USERNAME': '',
            'PASSWORD': '',
        }

    @staticmethod
    def django_app():
        return 'waldur_auth_bcc'

    @staticmethod
    def django_urls():
        from .urls import urlpatterns
        return urlpatterns



import django_filters
from django.db.models import Q

from waldur_core.core import filters as core_filters
from waldur_core.structure import filters as structure_filters

from . import models


class ProjectTemplateFilter(structure_filters.BaseServicePropertyFilter):
    class Meta(structure_filters.BaseServicePropertyFilter.Meta):
        model = models.ProjectTemplate


class ProjectFilter(structure_filters.BaseResourceFilter):
    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Project


class IssueTypeFilter(structure_filters.ServicePropertySettingsFilter):
    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.IssueType


class PriorityFilter(structure_filters.ServicePropertySettingsFilter):
    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.Priority


class IssueFilter(django_filters.FilterSet):
    created_before = django_filters.IsoDateTimeFilter(name="created", lookup_expr="lte")
    created_after = django_filters.IsoDateTimeFilter(name="created", lookup_expr="gte")
    summary = django_filters.CharFilter(lookup_expr='icontains')
    description = django_filters.CharFilter(lookup_expr='icontains')
    jira_project = core_filters.URLFilter(view_name='project-detail', name='project__uuid')
    jira_project_uuid = django_filters.UUIDFilter(name='project__uuid')
    priority_name = django_filters.ModelMultipleChoiceFilter(
        name='priority__name',
        to_field_name='name',
        queryset=models.Priority.objects.all()
    )
    project = core_filters.URLFilter(view_name='project-detail', name='project__service_project_link__project__uuid')
    project_uuid = django_filters.UUIDFilter(name='project__service_project_link__project__uuid')
    type_name = django_filters.CharFilter(name='type__name')
    updated_before = django_filters.IsoDateTimeFilter(name="updated", lookup_expr="lte")
    updated_after = django_filters.IsoDateTimeFilter(name="updated", lookup_expr="gte")
    user_uuid = django_filters.UUIDFilter(name='user__uuid')
    key = django_filters.CharFilter(name='backend_id')
    status = core_filters.LooseMultipleChoiceFilter()
    sla_ttr_breached = django_filters.BooleanFilter(name='resolution_sla', method='filter_resolution_sla',
                                                    widget=django_filters.widgets.BooleanWidget())

    def filter_resolution_sla(self, queryset, name, value):
        if value:
            return queryset.exclude(Q(resolution_sla__gte=0) | Q(resolution_sla__isnull=True))
        else:
            return queryset.filter(resolution_sla__gte=0)

    class Meta(object):
        model = models.Issue
        fields = [
            'description',
            'key',
            'status',
            'summary',
            'user_uuid',
            'creator_name',
            'assignee_name',
            'reporter_name',
        ]
        order_by = [
            'created',
            'updated',
            # desc
            '-created',
            '-updated',
        ]


class CommentFilter(django_filters.FilterSet):
    issue = core_filters.URLFilter(view_name='jira-issues-detail', name='issue__uuid')
    issue_uuid = django_filters.UUIDFilter(name='issue__uuid')
    user_uuid = django_filters.UUIDFilter(name='user__uuid')

    class Meta(object):
        model = models.Comment
        fields = []


class AttachmentFilter(django_filters.FilterSet):
    issue = core_filters.URLFilter(view_name='jira-issues-detail', name='issue__uuid')
    issue_uuid = django_filters.UUIDFilter(name='issue__uuid')

    class Meta(object):
        model = models.Attachment
        fields = []



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-10-02 07:24
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0038_drop_offerings'),
    ]

    operations = [
        migrations.CreateModel(
            name='OfferingComponent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('billing_type', models.CharField(choices=[('fixed', 'Fixed-price'), ('usage', 'Usage-based')], default='fixed', max_length=5)),
                ('type', models.CharField(help_text='Unique internal name of the measured unit, for example floating_ip.', max_length=50)),
                ('name', models.CharField(help_text='Display name for the measured unit, for example, Floating IP.', max_length=150)),
                ('measured_unit', models.CharField(help_text='Unit of measurement, for example, GB.', max_length=30, blank=True)),
                ('offering', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='components', to='marketplace.Offering')),
            ],
        ),
        migrations.AlterField(
            model_name='componentquota',
            name='component',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='marketplace.OfferingComponent'),
        ),
        migrations.AlterField(
            model_name='componentusage',
            name='component',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='marketplace.OfferingComponent'),
        ),
        migrations.AddField(
            model_name='plancomponent',
            name='component',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='components', to='marketplace.OfferingComponent'),
        ),
        migrations.AlterUniqueTogether(
            name='plancomponent',
            unique_together=set([('plan', 'component')]),
        ),
        migrations.RemoveField(
            model_name='plancomponent',
            name='billing_type',
        ),
        migrations.RemoveField(
            model_name='plancomponent',
            name='description',
        ),
        migrations.RemoveField(
            model_name='plancomponent',
            name='measured_unit',
        ),
        migrations.RemoveField(
            model_name='plancomponent',
            name='name',
        ),
        migrations.RemoveField(
            model_name='plancomponent',
            name='type',
        ),
        migrations.AlterUniqueTogether(
            name='offeringcomponent',
            unique_together=set([('type', 'offering')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-04-02 08:53
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0040_volume_type'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='volume',
            name='type',
        ),
        migrations.AddField(
            model_name='volume',
            name='type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    to='openstack_tenant.VolumeType'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-03-22 14:51
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('python_management', '0002_generalized_instance_added_system_user'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='pythonmanagement',
            unique_together=set([('instance_content_type', 'instance_object_id', 'virtual_envs_dir_path', 'system_user')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('support_invoices', '0002_delete_requestbasedoffering'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('support', '0036_offering_ordering'),
        ('invoices', '0030_json_details_on_genericinvoiceitem'),
        ('marketplace', '0078_fix_plan_component_amount'),
    ]

    operations = []



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('packages', '0003_add_meta_fields'),
    ]

    operations = [
        migrations.AddField(
            model_name='packagetemplate',
            name='category',
            field=models.CharField(default='small', max_length=10, choices=[('small', 'Small'), ('medium', 'Medium'), ('large', 'Large')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-01-22 14:51
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0052_resourceplanperiod'),
    ]

    operations = [
        migrations.CreateModel(
            name='CategoryComponent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(help_text='Display name for the measured unit, for example, Floating IP.', max_length=150)),
                ('type', models.CharField(help_text='Unique internal name of the measured unit, for example floating_ip.', max_length=50)),
                ('measured_unit', models.CharField(blank=True, help_text='Unit of measurement, for example, GB.', max_length=30)),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='components', to='marketplace.Category')),
            ],
        ),
        migrations.AddField(
            model_name='offeringcomponent',
            name='parent',
            field=models.ForeignKey(null=True, blank=True, on_delete=django.db.models.deletion.CASCADE, to='marketplace.CategoryComponent'),
        ),
        migrations.AlterUniqueTogether(
            name='categorycomponent',
            unique_together=set([('type', 'category')]),
        ),
    ]



from django.contrib import admin
from django.core.exceptions import ValidationError
from django.forms import ModelForm
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from waldur_core.core.admin import JsonWidget

from waldur_core.core.admin import ExecutorAdminAction, PasswordWidget
from waldur_core.quotas.admin import QuotaInline
from waldur_core.structure import admin as structure_admin

from . import executors, models


def _get_obj_admin_url(obj):
    return reverse('admin:%s_%s_change' % (obj._meta.app_label, obj._meta.model_name), args=[obj.id])


def _get_list_admin_url(model):
    return reverse('admin:%s_%s_changelist' % (model._meta.app_label, model._meta.model_name))


class ServiceProjectLinkAdmin(structure_admin.ServiceProjectLinkAdmin):
    readonly_fields = ('get_service_settings_username', 'get_service_settings_password') + \
        structure_admin.ServiceProjectLinkAdmin.readonly_fields

    def get_service_settings_username(self, obj):
        return obj.service.settings.username

    get_service_settings_username.short_description = _('Username')

    def get_service_settings_password(self, obj):
        return obj.service.settings.password

    get_service_settings_password.short_description = _('Password')


class TenantAdminForm(ModelForm):
    class Meta:
        widgets = {
            'extra_configuration': JsonWidget(),
            'user_password': PasswordWidget(),
        }


class TenantAdmin(structure_admin.ResourceAdmin):

    actions = ('pull', 'detect_external_networks', 'allocate_floating_ip', 'pull_security_groups',
               'pull_floating_ips', 'pull_quotas')
    inlines = [QuotaInline]
    form = TenantAdminForm

    class OKTenantAction(ExecutorAdminAction):
        """ Execute action with tenant that is in state OK """

        def validate(self, tenant):
            if tenant.state != models.Tenant.States.OK:
                raise ValidationError(_('Tenant has to be in state OK to pull security groups.'))

    class PullSecurityGroups(OKTenantAction):
        executor = executors.TenantPullSecurityGroupsExecutor
        short_description = _('Pull security groups')

    pull_security_groups = PullSecurityGroups()

    class AllocateFloatingIP(OKTenantAction):
        executor = executors.TenantAllocateFloatingIPExecutor
        short_description = _('Allocate floating IPs')

        def validate(self, tenant):
            super(TenantAdmin.AllocateFloatingIP, self).validate(tenant)
            if not tenant.external_network_id:
                raise ValidationError(_('Tenant has to have external network to allocate floating IP.'))

    allocate_floating_ip = AllocateFloatingIP()

    class DetectExternalNetworks(OKTenantAction):
        executor = executors.TenantDetectExternalNetworkExecutor
        short_description = _('Attempt to lookup and set external network id of the connected router')

    detect_external_networks = DetectExternalNetworks()

    class PullFloatingIPs(OKTenantAction):
        executor = executors.TenantPullFloatingIPsExecutor
        short_description = _('Pull floating IPs')

    pull_floating_ips = PullFloatingIPs()

    class PullQuotas(OKTenantAction):
        executor = executors.TenantPullQuotasExecutor
        short_description = _('Pull quotas')

    pull_quotas = PullQuotas()

    class Pull(ExecutorAdminAction):
        executor = executors.TenantPullExecutor
        short_description = _('Pull')

        def validate(self, tenant):
            if tenant.state not in (models.Tenant.States.OK, models.Tenant.States.ERRED):
                raise ValidationError(_('Tenant has to be OK or erred.'))
            if not tenant.backend_id:
                raise ValidationError(_('Tenant does not have backend ID.'))

    pull = Pull()


class FlavorAdmin(structure_admin.BackendModelAdmin):
    list_display = ('name', 'settings', 'cores', 'ram', 'disk')


class ImageAdmin(structure_admin.BackendModelAdmin):
    list_display = ('name', 'min_disk', 'min_ram')


class TenantResourceAdmin(structure_admin.ResourceAdmin):
    """ Admin model for resources that are connected to tenant.

        Expects that resource has attribute `tenant`.
    """
    list_display = structure_admin.ResourceAdmin.list_display + ('get_tenant', )
    list_filter = structure_admin.ResourceAdmin.list_filter + ('tenant', )

    def get_tenant(self, obj):
        tenant = obj.tenant
        return '<a href="%s">%s</a>' % (_get_obj_admin_url(tenant), tenant.name)

    get_tenant.short_description = _('Tenant')
    get_tenant.allow_tags = True


class NetworkAdmin(structure_admin.ResourceAdmin):
    list_display_links = None
    list_display = ('is_external', 'type') + structure_admin.ResourceAdmin.list_display
    fields = ('name', 'tenant', 'is_external', 'type', 'segmentation_id', 'state')


class SubNetAdmin(structure_admin.ResourceAdmin):
    list_display_links = None
    list_display = ('network', 'gateway_ip') + structure_admin.ResourceAdmin.list_display

    fields = ('name', 'network', 'cidr', 'gateway_ip', 'allocation_pools',
              'ip_version', 'enable_dhcp', 'dns_nameservers', 'state')


class CustomerOpenStackInline(admin.StackedInline):
    model = models.CustomerOpenStack
    extra = 1


admin.site.register(models.Network, NetworkAdmin)
admin.site.register(models.SubNet, SubNetAdmin)
admin.site.register(models.SecurityGroup, structure_admin.ResourceAdmin)

admin.site.register(models.Tenant, TenantAdmin)
admin.site.register(models.Flavor, FlavorAdmin)
admin.site.register(models.Image, ImageAdmin)
admin.site.register(models.OpenStackService, structure_admin.ServiceAdmin)
admin.site.register(models.OpenStackServiceProjectLink, ServiceProjectLinkAdmin)
admin.site.register(models.FloatingIP, structure_admin.ResourceAdmin)
structure_admin.CustomerAdmin.inlines += [CustomerOpenStackInline]



from __future__ import unicode_literals

import logging
import uuid

from django.conf import settings
from django_openid_auth.auth import OpenIDBackend


logger = logging.getLogger(__name__)


class WaldurOpenIDBackend(OpenIDBackend):
    """ This backend sets user's full_name and email. """

    def update_user_details(self, user, details, openid_response):
        updated_fields = []

        # Don't update full_name if it is already set
        if not user.full_name:
            user.full_name = '{} {}'.format(details['first_name'], details['last_name']).strip()
            updated_fields.append('full_name')

        # Don't update email if it is already set
        if not user.email and details['email']:
            user.email = details['email']
            updated_fields.append('email')

        # Civil number should be updated after each login because it can be changed or
        # defined for user.
        civil_number = self._get_civil_number(openid_response)
        if civil_number and user.civil_number != civil_number:
            user.civil_number = civil_number
            updated_fields.append('civil_number')

        if updated_fields:
            user.save(update_fields=updated_fields)

    def create_user_from_openid(self, openid_response):
        user = super(WaldurOpenIDBackend, self).create_user_from_openid(openid_response)
        civil_number = self._get_civil_number(openid_response)
        if civil_number and user.civil_number != civil_number:
            user.civil_number = civil_number
        user.registration_method = settings.WALDUR_AUTH_OPENID.get('NAME', 'openid')

        user.save()
        return user

    def _get_preferred_username(self, nickname, email):
        return uuid.uuid4().hex[:30]

    def _get_civil_number(self, openid_response):
        """
        Extract civil number from OpenID response.
        Return empty string if personal code is not defined.

        Expected openid.identity value: https://openid.ee/i/EE:<personal_code>
        Example: https://openid.ee/i/EE:37605030299
        Only the last part (<personal_code>) is stored as civil number.
        """
        openid_identity = openid_response.getSigned('http://specs.openid.net/auth/2.0', 'identity')

        personal_code = openid_identity.split('/')[-1].split(':')[-1]
        if personal_code.isdigit():
            return personal_code
        else:
            logger.warning(
                'Unable to parse openid.identity {}: personal code is not a numeric value'.format(openid_identity))
            return ''



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-30 15:37
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0039_immutable_default_json'),
    ]

    operations = [
        migrations.AlterField(
            model_name='floatingip',
            name='address',
            field=models.GenericIPAddressField(blank=True, default=None, null=True, protocol='IPv4'),
        ),
        migrations.AlterUniqueTogether(
            name='floatingip',
            unique_together=set([('tenant', 'address')]),
        ),
    ]



from django import template

from waldur_mastermind.marketplace import plugins

from .. import models

register = template.Library()


@register.simple_tag
def get_invoice_item_component_amount(item, component):
    available_limits = plugins.manager.get_available_limits(component.component.offering.type)
    if item.limits and (
        component.component.billing_type == models.OfferingComponent.BillingTypes.USAGE or
        component.component.type in available_limits
    ):
        builtin_components = plugins.manager.get_components(component.component.offering.type)
        component_factors = {c.type: c.factor for c in builtin_components}
        factor = component_factors.get(component.component.type, 1)
        limit = item.limits.get(component.component.type, 0)
        return limit / factor

    return component.amount



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-12-04 09:50
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0003_order_by_name'),
    ]

    operations = [
        migrations.AddField(
            model_name='customer',
            name='homepage',
            field=models.URLField(blank=True, max_length=255),
        ),
    ]



import django_filters

from waldur_core.core import filters as core_filters
from waldur_core.structure import filters as structure_filters

from . import models


class OpenStackServiceProjectLinkFilter(structure_filters.BaseServiceProjectLinkFilter):
    service = core_filters.URLFilter(view_name='openstack-detail', name='service__uuid')

    class Meta(structure_filters.BaseServiceProjectLinkFilter.Meta):
        model = models.OpenStackServiceProjectLink


class SecurityGroupFilter(structure_filters.BaseResourceFilter):
    tenant_uuid = django_filters.UUIDFilter(name='tenant__uuid')
    tenant = core_filters.URLFilter(view_name='openstack-tenant-detail', name='tenant__uuid')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.SecurityGroup


class FloatingIPFilter(structure_filters.BaseResourceFilter):
    tenant_uuid = django_filters.UUIDFilter(name='tenant__uuid')
    tenant = core_filters.URLFilter(view_name='openstack-tenant-detail', name='tenant__uuid')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.FloatingIP
        fields = structure_filters.BaseResourceFilter.Meta.fields + ('runtime_state', )


class FlavorFilter(structure_filters.ServicePropertySettingsFilter):

    o = django_filters.OrderingFilter(fields=('cores', 'ram', 'disk'))

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.Flavor
        fields = dict({
            'cores': ['exact', 'gte', 'lte'],
            'ram': ['exact', 'gte', 'lte'],
            'disk': ['exact', 'gte', 'lte'],
        }, **{field: ['exact'] for field in structure_filters.ServicePropertySettingsFilter.Meta.fields})


class ImageFilter(structure_filters.ServicePropertySettingsFilter):

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.Image


class NetworkFilter(structure_filters.BaseResourceFilter):
    tenant_uuid = django_filters.UUIDFilter(name='tenant__uuid')
    tenant = core_filters.URLFilter(view_name='openstack-tenant-detail', name='tenant__uuid')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Network


class SubNetFilter(structure_filters.BaseResourceFilter):
    tenant_uuid = django_filters.UUIDFilter(name='network__tenant__uuid')
    tenant = core_filters.URLFilter(view_name='openstack-tenant-detail', name='network__tenant__uuid')
    network_uuid = django_filters.UUIDFilter(name='network__uuid')
    network = core_filters.URLFilter(view_name='openstack-network-detail', name='network__uuid')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.SubNet
